__all__ = (
    "setup_dither_stackers",
    "wrap_ra_dec",
    "wrap_ra",
    "in_hexagon",
    "polygon_coords",
    "BaseDitherStacker",
    "RandomDitherFieldPerVisitStacker",
    "RandomDitherFieldPerNightStacker",
    "SpiralDitherFieldPerVisitStacker",
    "SpiralDitherFieldPerNightStacker",
    "HexDitherFieldPerVisitStacker",
    "HexDitherFieldPerNightStacker",
    "RandomDitherPerNightStacker",
    "SpiralDitherPerNightStacker",
    "HexDitherPerNightStacker",
    "RandomRotDitherPerFilterChangeStacker",
)


import warnings

import numpy as np

from .base_stacker import BaseStacker

# Stacker naming scheme:
# [Pattern]Dither[Field]Per[Timescale].
#  Timescale indicates how often the dither offset is changed.
#  The presence of 'Field' indicates that a new offset is chosen per field, on the indicated timescale.
#  The absence of 'Field' indicates that all visits within the indicated timescale use the same dither offset.


# Original dither stackers (Random, Spiral, Hex) written by Lynne Jones (lynnej@uw.edu)
# Additional dither stackers written by Humna Awan (humna.awan@rutgers.edu), with addition of
# constraining dither offsets to be within an inscribed hexagon (code modifications for use here by LJ).


def setup_dither_stackers(ra_col, dec_col, degrees, **kwargs):
    b = BaseStacker()
    stacker_list = []
    if ra_col in b.source_dict:
        stacker_list.append(b.source_dict[ra_col](degrees=degrees, **kwargs))
    if dec_col in b.source_dict:
        if b.source_dict[ra_col] != b.source_dict[dec_col]:
            stacker_list.append(b.source_dict[dec_col](degrees=degrees, **kwargs))
    return stacker_list


def wrap_ra_dec(ra, dec):
    """
    Wrap RA into 0-2pi and Dec into +/0 pi/2.

    Parameters
    ----------
    ra : numpy.ndarray
        RA in radians
    dec : numpy.ndarray
        Dec in radians

    Returns
    -------
    numpy.ndarray, numpy.ndarray
        Wrapped RA/Dec values, in radians.
    """
    # Wrap dec.
    low = np.where(dec < -np.pi / 2.0)[0]
    dec[low] = -1 * (np.pi + dec[low])
    ra[low] = ra[low] - np.pi
    high = np.where(dec > np.pi / 2.0)[0]
    dec[high] = np.pi - dec[high]
    ra[high] = ra[high] - np.pi
    # Wrap RA.
    ra = ra % (2.0 * np.pi)
    return ra, dec


def wrap_ra(ra):
    """
    Wrap only RA values into 0-2pi (using mod).

    Parameters
    ----------
    ra : numpy.ndarray
        RA in radians

    Returns
    -------
    numpy.ndarray
        Wrapped RA values, in radians.
    """
    ra = ra % (2.0 * np.pi)
    return ra


def in_hexagon(x_off, y_off, max_dither):
    """
    Identify dither offsets which fall within the inscribed hexagon.

    Parameters
    ----------
    x_off : numpy.ndarray
        The x values of the dither offsets.
    yoff : numpy.ndarray
        The y values of the dither offsets.
    max_dither : float
        The maximum dither offset.

    Returns
    -------
    numpy.ndarray
        Indexes of the offsets which are within the hexagon inscribed inside the 'max_dither' radius circle.
    """
    # Set up the hexagon limits.
    #  y = mx + b, 2h is the height.
    m = np.sqrt(3.0)
    b = m * max_dither
    h = m / 2.0 * max_dither
    # Identify offsets inside hexagon.
    inside = np.where(
        (y_off < m * x_off + b)
        & (y_off > m * x_off - b)
        & (y_off < -m * x_off + b)
        & (y_off > -m * x_off - b)
        & (y_off < h)
        & (y_off > -h)
    )[0]
    return inside


def polygon_coords(nside, radius, rotation_angle):
    """
    Find the x,y coords of a polygon.

    This is useful for plotting dither points and showing they lie within
    a given shape.

    Parameters
    ----------
    nside : int
        The number of sides of the polygon
    radius : float
        The radius within which to plot the polygon
    rotation_angle : float
        The angle to rotate the polygon to.

    Returns
    -------
    [float, float]
        List of x/y coordinates of the points describing the polygon.
    """
    each_angle = 2 * np.pi / float(nside)
    x_coords = np.zeros(nside, float)
    y_coords = np.zeros(nside, float)
    for i in range(0, nside):
        x_coords[i] = np.sin(each_angle * i + rotation_angle) * radius
        y_coords[i] = np.cos(each_angle * i + rotation_angle) * radius
    return list(zip(x_coords, y_coords))


class BaseDitherStacker(BaseStacker):
    """Base class for dither stackers.

    The base class just adds an easy way to define a stacker as one of the 'dither' types of stackers.
    These run first, before any other stackers.

    Parameters
    ----------
    ra_col : str, optional
        The name of the RA column in the data.
        Default 'fieldRA'.
    dec_col : str, optional
        The name of the Dec column in the data.
        Default 'fieldDec'.
    degrees : bool, optional
        Flag whether RA/Dec should be treated as (and kept as) degrees.
    max_dither : float, optional
        The radius of the maximum dither offset, in degrees.
        Default 1.75 degrees.
    in_hex : bool, optional
        If True, offsets are constrained to lie within a hexagon inscribed within the max_dither circle.
        If False, offsets can lie anywhere out to the edges of the max_dither circle.
        Default True.
    """

    cols_added = []

    def __init__(
        self,
        ra_col="fieldRA",
        dec_col="fieldDec",
        degrees=True,
        max_dither=1.75,
        in_hex=True,
    ):
        # Instantiate the RandomDither object and set internal variables.
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.degrees = degrees
        # Convert max_dither to radians for internal use.
        self.max_dither = np.radians(max_dither)
        self.in_hex = in_hex
        # self.units used for plot labels
        if self.degrees:
            self.units = ["deg", "deg"]
        else:
            self.units = ["rad", "rad"]
        # Values required for framework operation: this specifies the data columns required from the database.
        self.cols_req = [self.ra_col, self.dec_col]


class RandomDitherFieldPerVisitStacker(BaseDitherStacker):
    """
    Randomly dither the RA and Dec pointings up to max_dither degrees from center,
    with a different offset for each field, for each visit.

    Parameters
    ----------
    ra_col : str, optional
        The name of the RA column in the data.
        Default 'fieldRA'.
    dec_col : str, optional
        The name of the Dec column in the data.
        Default 'fieldDec'.
    degrees : bool, optional
        Flag whether RA/Dec should be treated as (and kept as) degrees.
    max_dither : float, optional
        The radius of the maximum dither offset, in degrees.
        Default 1.75 degrees.
    in_hex : bool, optional
        If True, offsets are constrained to lie within a hexagon inscribed within the max_dither circle.
        If False, offsets can lie anywhere out to the edges of the max_dither circle.
        Default True.
    random_seed : int or None, optional
        If set, then used as the random seed for the numpy random number generation for the dither offsets.
        Default None.
    """

    # Values required for framework operation: this specifies the name of the new columns.
    cols_added = ["randomDitherFieldPerVisitRa", "randomDitherFieldPerVisitDec"]

    def __init__(
        self,
        ra_col="fieldRA",
        dec_col="fieldDec",
        degrees=True,
        max_dither=1.75,
        in_hex=True,
        random_seed=None,
    ):
        """
        @ MaxDither in degrees
        """
        super().__init__(
            ra_col=ra_col,
            dec_col=dec_col,
            degrees=degrees,
            max_dither=max_dither,
            in_hex=in_hex,
        )
        self.random_seed = random_seed

    def _generate_random_offsets(self, noffsets):
        x_out = np.array([], float)
        y_out = np.array([], float)
        max_tries = 100
        tries = 0
        while (len(x_out) < noffsets) and (tries < max_tries):
            dithers_rad = np.sqrt(self._rng.rand(noffsets * 2)) * self.max_dither
            dithers_theta = self._rng.rand(noffsets * 2) * np.pi * 2.0
            x_off = dithers_rad * np.cos(dithers_theta)
            y_off = dithers_rad * np.sin(dithers_theta)
            if self.in_hex:
                # Constrain dither offsets to be within hexagon.
                idx = in_hexagon(x_off, y_off, self.max_dither)
                x_off = x_off[idx]
                y_off = y_off[idx]
            x_out = np.concatenate([x_out, x_off])
            y_out = np.concatenate([y_out, y_off])
            tries += 1
        if len(x_out) < noffsets:
            raise ValueError(
                "Could not find enough random points within the hexagon in %d tries. "
                "Try another random seed?" % (max_tries)
            )
        self.x_off = x_out[0:noffsets]
        self.y_off = y_out[0:noffsets]

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Column already present in data; assume it is correct and does not need recalculating.
            return sim_data
        # Generate random numbers for dither, using defined seed value if desired.
        if not hasattr(self, "_rng"):
            if self.random_seed is not None:
                self._rng = np.random.RandomState(self.random_seed)
            else:
                self._rng = np.random.RandomState(2178813)

        # Generate the random dither values.
        noffsets = len(sim_data[self.ra_col])
        self._generate_random_offsets(noffsets)
        # Add to RA and dec values.
        if self.degrees:
            ra = np.radians(sim_data[self.ra_col])
            dec = np.radians(sim_data[self.dec_col])
        else:
            ra = sim_data[self.ra_col]
            dec = sim_data[self.dec_col]
        sim_data["randomDitherFieldPerVisitRa"] = ra + self.x_off / np.cos(dec)
        sim_data["randomDitherFieldPerVisitDec"] = dec + self.y_off
        # Wrap back into expected range.
        (
            sim_data["randomDitherFieldPerVisitRa"],
            sim_data["randomDitherFieldPerVisitDec"],
        ) = wrap_ra_dec(
            sim_data["randomDitherFieldPerVisitRa"],
            sim_data["randomDitherFieldPerVisitDec"],
        )
        # Convert to degrees
        if self.degrees:
            for col in self.cols_added:
                sim_data[col] = np.degrees(sim_data[col])
        return sim_data


class RandomDitherFieldPerNightStacker(RandomDitherFieldPerVisitStacker):
    """
    Randomly dither the RA and Dec pointings up to max_dither degrees from center,
    one dither offset per new night of observation of a field.
    e.g. visits within the same night, to the same field, have the same offset.

    Parameters
    ----------
    ra_col : str, optional
        The name of the RA column in the data.
        Default 'fieldRA'.
    dec_col : str, optional
        The name of the Dec column in the data.
        Default 'fieldDec'.
    degrees : bool, optional
        Flag whether RA/Dec should be treated as (and kept as) degrees.
    field_id_col : str, optional
        The name of the fieldId column in the data.
        Used to identify fields which should be identified as the 'same'.
        Default 'fieldId'.
    night_col : str, optional
        The name of the night column in the data.
        Default 'night'.
    max_dither : float, optional
        The radius of the maximum dither offset, in degrees.
        Default 1.75 degrees.
    in_hex : bool, optional
        If True, offsets are constrained to lie within a hexagon inscribed within the max_dither circle.
        If False, offsets can lie anywhere out to the edges of the max_dither circle.
        Default True.
    random_seed : int or None, optional
        If set, then used as the random seed for the numpy random number generation for the dither offsets.
        Default None.
    """

    # Values required for framework operation: this specifies the names of the new columns.
    cols_added = ["randomDitherFieldPerNightRa", "randomDitherFieldPerNightDec"]

    def __init__(
        self,
        ra_col="fieldRA",
        dec_col="fieldDec",
        degrees=True,
        field_id_col="fieldId",
        night_col="night",
        max_dither=1.75,
        in_hex=True,
        random_seed=None,
    ):
        """
        @ MaxDither in degrees
        """
        # Instantiate the RandomDither object and set internal variables.
        super().__init__(
            ra_col=ra_col,
            dec_col=dec_col,
            degrees=degrees,
            max_dither=max_dither,
            in_hex=in_hex,
            random_seed=random_seed,
        )
        self.night_col = night_col
        self.field_id_col = field_id_col
        # Values required for framework operation: this specifies the data columns required from the database.
        self.cols_req = [self.ra_col, self.dec_col, self.night_col, self.field_id_col]

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            return sim_data
        # Generate random numbers for dither, using defined seed value if desired.
        if not hasattr(self, "_rng"):
            if self.random_seed is not None:
                self._rng = np.random.RandomState(self.random_seed)
            else:
                self._rng = np.random.RandomState(872453)

        # Generate the random dither values, one per night per field.
        fields = np.unique(sim_data[self.field_id_col])
        nights = np.unique(sim_data[self.night_col])
        self._generate_random_offsets(len(fields) * len(nights))
        if self.degrees:
            ra = np.radians(sim_data[self.ra_col])
            dec = np.radians(sim_data[self.dec_col])
        else:
            ra = sim_data[self.ra_col]
            dec = sim_data[self.dec_col]
        # counter to ensure new random numbers are chosen every time
        delta = 0
        for fieldid in np.unique(sim_data[self.field_id_col]):
            # Identify observations of this field.
            match = np.where(sim_data[self.field_id_col] == fieldid)[0]
            # Apply dithers, increasing each night.
            nights = sim_data[self.night_col][match]
            vertex_idxs = np.searchsorted(np.unique(nights), nights)
            vertex_idxs = vertex_idxs % len(self.x_off)
            # ensure that the same xOff/yOff entries are not chosen
            delta = delta + len(vertex_idxs)
            sim_data["randomDitherFieldPerNightRa"][match] = ra[match] + self.x_off[vertex_idxs] / np.cos(
                dec[match]
            )
            sim_data["randomDitherFieldPerNightDec"][match] = dec[match] + self.y_off[vertex_idxs]
        # Wrap into expected range.
        (
            sim_data["randomDitherFieldPerNightRa"],
            sim_data["randomDitherFieldPerNightDec"],
        ) = wrap_ra_dec(
            sim_data["randomDitherFieldPerNightRa"],
            sim_data["randomDitherFieldPerNightDec"],
        )
        if self.degrees:
            for col in self.cols_added:
                sim_data[col] = np.degrees(sim_data[col])
        return sim_data


class RandomDitherPerNightStacker(RandomDitherFieldPerVisitStacker):
    """
    Randomly dither the RA and Dec pointings up to max_dither degrees from center,
    one dither offset per night.
    All fields observed within the same night get the same offset.

    Parameters
    ----------
    ra_col : str, optional
        The name of the RA column in the data.
        Default 'fieldRA'.
    dec_col : str, optional
        The name of the Dec column in the data.
        Default 'fieldDec'.
    degrees : bool, optional
        Flag whether RA/Dec should be treated as (and kept as) degrees.
    night_col : str, optional
        The name of the night column in the data.
        Default 'night'.
    max_dither : float, optional
        The radius of the maximum dither offset, in degrees.
        Default 1.75 degrees.
    in_hex : bool, optional
        If True, offsets are constrained to lie within a hexagon inscribed within the max_dither circle.
        If False, offsets can lie anywhere out to the edges of the max_dither circle.
        Default True.
    random_seed : int or None, optional
        If set, then used as the random seed for the numpy random number generation for the dither offsets.
        Default None.
    """

    # Values required for framework operation: this specifies the names of the new columns.
    cols_added = ["randomDitherPerNightRa", "randomDitherPerNightDec"]

    def __init__(
        self,
        ra_col="fieldRA",
        dec_col="fieldDec",
        degrees=True,
        night_col="night",
        max_dither=1.75,
        in_hex=True,
        random_seed=None,
    ):
        """
        @ MaxDither in degrees
        """
        # Instantiate the RandomDither object and set internal variables.
        super().__init__(
            ra_col=ra_col,
            dec_col=dec_col,
            degrees=degrees,
            max_dither=max_dither,
            in_hex=in_hex,
            random_seed=random_seed,
        )
        self.night_col = night_col
        # Values required for framework operation: this specifies the data columns required from the database.
        self.cols_req = [self.ra_col, self.dec_col, self.night_col]

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            return sim_data
        # Generate random numbers for dither, using defined seed value if desired.
        if not hasattr(self, "_rng"):
            if self.random_seed is not None:
                self._rng = np.random.RandomState(self.random_seed)
            else:
                self._rng = np.random.RandomState(66334)

        # Generate the random dither values, one per night.
        nights = np.unique(sim_data[self.night_col])
        self._generate_random_offsets(len(nights))
        if self.degrees:
            ra = np.radians(sim_data[self.ra_col])
            dec = np.radians(sim_data[self.dec_col])
        else:
            ra = sim_data[self.ra_col]
            dec = sim_data[self.dec_col]
        # Add to RA and dec values.
        for n, x, y in zip(nights, self.x_off, self.y_off):
            match = np.where(sim_data[self.night_col] == n)[0]
            sim_data["randomDitherPerNightRa"][match] = ra[match] + x / np.cos(dec[match])
            sim_data["randomDitherPerNightDec"][match] = dec[match] + y
        # Wrap RA/Dec into expected range.
        (
            sim_data["randomDitherPerNightRa"],
            sim_data["randomDitherPerNightDec"],
        ) = wrap_ra_dec(sim_data["randomDitherPerNightRa"], sim_data["randomDitherPerNightDec"])
        if self.degrees:
            for col in self.cols_added:
                sim_data[col] = np.degrees(sim_data[col])
        return sim_data


class SpiralDitherFieldPerVisitStacker(BaseDitherStacker):
    """
    Offset along an equidistant spiral with num_points, out to a maximum radius of max_dither.
    Each visit to a field receives a new, sequential offset.

    Parameters
    ----------
    ra_col : str, optional
        The name of the RA column in the data.
        Default 'fieldRA'.
    dec_col : str, optional
        The name of the Dec column in the data.
        Default 'fieldDec'.
    degrees : bool, optional
        Flag whether RA/Dec should be treated as (and kept as) degrees.
    field_id_col : str, optional
        The name of the fieldId column in the data.
        Used to identify fields which should be identified as the 'same'.
        Default 'fieldId'.
    num_points : int, optional
        The number of points in the spiral.
        Default 60.
    max_dither : float, optional
        The radius of the maximum dither offset, in degrees.
        Default 1.75 degrees.
    n_coils : int, optional
        The number of coils the spiral should have.
        Default 5.
    in_hex : bool, optional
        If True, offsets are constrained to lie within a hexagon inscribed within the max_dither circle.
        If False, offsets can lie anywhere out to the edges of the max_dither circle.
        Default True.
    """

    # Values required for framework operation: this specifies the names of the new columns.
    cols_added = ["spiralDitherFieldPerVisitRa", "spiralDitherFieldPerVisitDec"]

    def __init__(
        self,
        ra_col="fieldRA",
        dec_col="fieldDec",
        degrees=True,
        field_id_col="fieldId",
        num_points=60,
        max_dither=1.75,
        n_coils=5,
        in_hex=True,
    ):
        """
        @ MaxDither in degrees
        """
        super().__init__(
            ra_col=ra_col,
            dec_col=dec_col,
            degrees=degrees,
            max_dither=max_dither,
            in_hex=in_hex,
        )
        self.field_id_col = field_id_col
        # Convert max_dither from degrees (internal units for ra/dec are radians)
        self.num_points = num_points
        self.n_coils = n_coils
        # Values required for framework operation: this specifies the data columns required from the database.
        self.cols_req = [self.ra_col, self.dec_col, self.field_id_col]

    def _generate_spiral_offsets(self):
        # First generate a full archimedean spiral ..
        theta = np.arange(0.0001, self.n_coils * np.pi * 2.0, 0.001)
        a = self.max_dither / theta.max()
        if self.in_hex:
            a = 0.85 * a
        r = theta * a
        # Then pick out equidistant points along the spiral.
        arc = a / 2.0 * (theta * np.sqrt(1 + theta**2) + np.log(theta + np.sqrt(1 + theta**2)))
        stepsize = arc.max() / float(self.num_points)
        arcpts = np.arange(0, arc.max(), stepsize)
        arcpts = arcpts[0 : self.num_points]
        rpts = np.zeros(self.num_points, float)
        thetapts = np.zeros(self.num_points, float)
        for i, ap in enumerate(arcpts):
            diff = np.abs(arc - ap)
            match = np.where(diff == diff.min())[0][0]
            rpts[i] = r[match]
            thetapts[i] = theta[match]
        # Translate these r/theta points into x/y (ra/dec) offsets.
        self.x_off = rpts * np.cos(thetapts)
        self.y_off = rpts * np.sin(thetapts)

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            return sim_data
        # Generate the spiral offset vertices.
        self._generate_spiral_offsets()
        # Now apply to observations.
        if self.degrees:
            ra = np.radians(sim_data[self.ra_col])
            dec = np.radians(sim_data[self.dec_col])
        else:
            ra = sim_data[self.ra_col]
            dec = sim_data[self.dec_col]
        for fieldid in np.unique(sim_data[self.field_id_col]):
            match = np.where(sim_data[self.field_id_col] == fieldid)[0]
            # Apply sequential dithers, increasing with each visit.
            vertex_idxs = np.arange(0, len(match), 1)
            vertex_idxs = vertex_idxs % self.num_points
            sim_data["spiralDitherFieldPerVisitRa"][match] = ra[match] + self.x_off[vertex_idxs] / np.cos(
                dec[match]
            )
            sim_data["spiralDitherFieldPerVisitDec"][match] = dec[match] + self.y_off[vertex_idxs]
        # Wrap into expected range.
        (
            sim_data["spiralDitherFieldPerVisitRa"],
            sim_data["spiralDitherFieldPerVisitDec"],
        ) = wrap_ra_dec(
            sim_data["spiralDitherFieldPerVisitRa"],
            sim_data["spiralDitherFieldPerVisitDec"],
        )
        if self.degrees:
            for col in self.cols_added:
                sim_data[col] = np.degrees(sim_data[col])
        return sim_data


class SpiralDitherFieldPerNightStacker(SpiralDitherFieldPerVisitStacker):
    """
    Offset along an equidistant spiral with num_points, out to a maximum radius of max_dither.
    Each field steps along a sequential series of offsets, each night it is observed.

    Parameters
    ----------
    ra_col : str, optional
        The name of the RA column in the data.
        Default 'fieldRA'.
    dec_col : str, optional
        The name of the Dec column in the data.
        Default 'fieldDec'.
    degrees : bool, optional
        Flag whether RA/Dec should be treated as (and kept as) degrees.
    field_id_col : str, optional
        The name of the fieldId column in the data.
        Used to identify fields which should be identified as the 'same'.
        Default 'fieldId'.
    night_col : str, optional
        The name of the night column in the data.
        Default 'night'.
    num_points : int, optional
        The number of points in the spiral.
        Default 60.
    max_dither : float, optional
        The radius of the maximum dither offset, in degrees.
        Default 1.75 degrees.
    n_coils : int, optional
        The number of coils the spiral should have.
        Default 5.
    in_hex : bool, optional
        If True, offsets are constrained to lie within a hexagon inscribed within the max_dither circle.
        If False, offsets can lie anywhere out to the edges of the max_dither circle.
        Default True.
    """

    # Values required for framework operation: this specifies the names of the new columns.
    cols_added = ["spiralDitherFieldPerNightRa", "spiralDitherFieldPerNightDec"]

    def __init__(
        self,
        ra_col="fieldRA",
        dec_col="fieldDec",
        degrees=True,
        field_id_col="fieldId",
        night_col="night",
        num_points=60,
        max_dither=1.75,
        n_coils=5,
        in_hex=True,
    ):
        """
        @ MaxDither in degrees
        """
        super().__init__(
            ra_col=ra_col,
            dec_col=dec_col,
            degrees=degrees,
            field_id_col=field_id_col,
            num_points=num_points,
            max_dither=max_dither,
            n_coils=n_coils,
            in_hex=in_hex,
        )
        self.night_col = night_col
        # Values required for framework operation: this specifies the data columns required from the database.
        self.cols_req.append(self.night_col)

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            return sim_data
        self._generate_spiral_offsets()
        if self.degrees:
            ra = np.radians(sim_data[self.ra_col])
            dec = np.radians(sim_data[self.dec_col])
        else:
            ra = sim_data[self.ra_col]
            dec = sim_data[self.dec_col]
        for fieldid in np.unique(sim_data[self.field_id_col]):
            # Identify observations of this field.
            match = np.where(sim_data[self.field_id_col] == fieldid)[0]
            # Apply a sequential dither, increasing each night.
            nights = sim_data[self.night_col][match]
            vertex_idxs = np.searchsorted(np.unique(nights), nights)
            vertex_idxs = vertex_idxs % self.num_points
            sim_data["spiralDitherFieldPerNightRa"][match] = ra[match] + self.x_off[vertex_idxs] / np.cos(
                dec[match]
            )
            sim_data["spiralDitherFieldPerNightDec"][match] = dec[match] + self.y_off[vertex_idxs]
        # Wrap into expected range.
        (
            sim_data["spiralDitherFieldPerNightRa"],
            sim_data["spiralDitherFieldPerNightDec"],
        ) = wrap_ra_dec(
            sim_data["spiralDitherFieldPerNightRa"],
            sim_data["spiralDitherFieldPerNightDec"],
        )
        if self.degrees:
            for col in self.cols_added:
                sim_data[col] = np.degrees(sim_data[col])
        return sim_data


class SpiralDitherPerNightStacker(SpiralDitherFieldPerVisitStacker):
    """
    Offset along an equidistant spiral with num_points, out to a maximum radius of max_dither.
    All fields observed in the same night receive the same sequential offset, changing per night.

    Parameters
    ----------
    ra_col : str, optional
        The name of the RA column in the data.
        Default 'fieldRA'.
    dec_col : str, optional
        The name of the Dec column in the data.
        Default 'fieldDec'.
    degrees : bool, optional
        Flag whether RA/Dec should be treated as (and kept as) degrees.
    field_id_col : str, optional
        The name of the fieldId column in the data.
        Used to identify fields which should be identified as the 'same'.
        Default 'fieldId'.
    night_col : str, optional
        The name of the night column in the data.
        Default 'night'.
    num_points : int, optional
        The number of points in the spiral.
        Default 60.
    max_dither : float, optional
        The radius of the maximum dither offset, in degrees.
        Default 1.75 degrees.
    n_coils : int, optional
        The number of coils the spiral should have.
        Default 5.
    in_hex : bool, optional
        If True, offsets are constrained to lie within a hexagon inscribed within the max_dither circle.
        If False, offsets can lie anywhere out to the edges of the max_dither circle.
        Default True.
    """

    # Values required for framework operation: this specifies the names of the new columns.
    cols_added = ["spiralDitherPerNightRa", "spiralDitherPerNightDec"]

    def __init__(
        self,
        ra_col="fieldRA",
        dec_col="fieldDec",
        degrees=True,
        field_id_col="fieldId",
        night_col="night",
        num_points=60,
        max_dither=1.75,
        n_coils=5,
        in_hex=True,
    ):
        """
        @ MaxDither in degrees
        """
        super().__init__(
            ra_col=ra_col,
            dec_col=dec_col,
            degrees=degrees,
            field_id_col=field_id_col,
            num_points=num_points,
            max_dither=max_dither,
            n_coils=n_coils,
            in_hex=in_hex,
        )
        self.night_col = night_col
        # Values required for framework operation: this specifies the data columns required from the database.
        self.cols_req.append(self.night_col)

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            return sim_data
        self._generate_spiral_offsets()
        nights = np.unique(sim_data[self.night_col])
        if self.degrees:
            ra = np.radians(sim_data[self.ra_col])
            dec = np.radians(sim_data[self.dec_col])
        else:
            ra = sim_data[self.ra_col]
            dec = sim_data[self.dec_col]
        # Add to RA and dec values.
        vertex_idxs = np.searchsorted(nights, sim_data[self.night_col])
        vertex_idxs = vertex_idxs % self.num_points
        sim_data["spiralDitherPerNightRa"] = ra + self.x_off[vertex_idxs] / np.cos(dec)
        sim_data["spiralDitherPerNightDec"] = dec + self.y_off[vertex_idxs]
        # Wrap RA/Dec into expected range.
        (
            sim_data["spiralDitherPerNightRa"],
            sim_data["spiralDitherPerNightDec"],
        ) = wrap_ra_dec(sim_data["spiralDitherPerNightRa"], sim_data["spiralDitherPerNightDec"])
        if self.degrees:
            for col in self.cols_added:
                sim_data[col] = np.degrees(sim_data[col])
        return sim_data


class HexDitherFieldPerVisitStacker(BaseDitherStacker):
    """
    Use offsets from the hexagonal grid of 'hexdither', but visit each vertex sequentially.
    Sequential offset for each visit.

    Parameters
    ----------
    ra_col : str, optional
        The name of the RA column in the data.
        Default 'fieldRA'.
    dec_col : str, optional
        The name of the Dec column in the data.
        Default 'fieldDec'.
    degrees : bool, optional
        Flag whether RA/Dec should be treated as (and kept as) degrees.
    field_id_col : str, optional
        The name of the fieldId column in the data.
        Used to identify fields which should be identified as the 'same'.
        Default 'fieldId'.
    max_dither : float, optional
        The radius of the maximum dither offset, in degrees.
        Default 1.75 degrees.
    in_hex : bool, optional
        If True, offsets are constrained to lie within a hexagon inscribed within the max_dither circle.
        If False, offsets can lie anywhere out to the edges of the max_dither circle.
        Default True.
    """

    # Values required for framework operation: this specifies the names of the new columns.
    cols_added = ["hexDitherFieldPerVisitRa", "hexDitherFieldPerVisitDec"]

    def __init__(
        self,
        ra_col="fieldRA",
        dec_col="fieldDec",
        degrees=True,
        field_id_col="fieldId",
        max_dither=1.75,
        in_hex=True,
    ):
        """
        @ MaxDither in degrees
        """
        super().__init__(
            ra_col=ra_col,
            dec_col=dec_col,
            degrees=degrees,
            max_dither=max_dither,
            in_hex=in_hex,
        )
        self.field_id_col = field_id_col
        # Values required for framework operation: this specifies the data columns required from the database.
        self.cols_req = [self.ra_col, self.dec_col, self.field_id_col]

    def _generate_hex_offsets(self):
        # Set up basics of dither pattern.
        dith_level = 4
        nrows = 2**dith_level
        halfrows = int(nrows / 2.0)
        # Calculate size of each offset
        dith_size_x = self.max_dither * 2.0 / float(nrows)
        dith_size_y = np.sqrt(3) * self.max_dither / float(nrows)  # sqrt 3 comes from hexagon
        if self.in_hex:
            dith_size_x = 0.95 * dith_size_x
            dith_size_y = 0.95 * dith_size_y
        # Calculate the row identification number, going from 0 at center
        nid_row = np.arange(-halfrows, halfrows + 1, 1)
        # and calculate the number of vertices in each row.
        vert_in_row = np.arange(-halfrows, halfrows + 1, 1)
        # First calculate how many vertices we will create in each row.
        total_vert = 0
        for i in range(-halfrows, halfrows + 1, 1):
            vert_in_row[i] = (nrows + 1) - abs(nid_row[i])
            total_vert += vert_in_row[i]
        self.num_points = total_vert
        self.x_off = []
        self.y_off = []
        # Calculate offsets over hexagonal grid.
        for i in range(0, nrows + 1, 1):
            for j in range(0, vert_in_row[i], 1):
                self.x_off.append(dith_size_x * (j - (vert_in_row[i] - 1) / 2.0))
                self.y_off.append(dith_size_y * nid_row[i])
        self.x_off = np.array(self.x_off)
        self.y_off = np.array(self.y_off)

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            return sim_data
        self._generate_hex_offsets()
        if self.degrees:
            ra = np.radians(sim_data[self.ra_col])
            dec = np.radians(sim_data[self.dec_col])
        else:
            ra = sim_data[self.ra_col]
            dec = sim_data[self.dec_col]
        for fieldid in np.unique(sim_data[self.field_id_col]):
            # Identify observations of this field.
            match = np.where(sim_data[self.field_id_col] == fieldid)[0]
            # Apply sequential dithers, increasing with each visit.
            vertex_idxs = np.arange(0, len(match), 1)
            vertex_idxs = vertex_idxs % self.num_points
            sim_data["hexDitherFieldPerVisitRa"][match] = ra[match] + self.x_off[vertex_idxs] / np.cos(
                dec[match]
            )
            sim_data["hexDitherFieldPerVisitDec"][match] = dec[match] + self.y_off[vertex_idxs]
        # Wrap into expected range.
        (
            sim_data["hexDitherFieldPerVisitRa"],
            sim_data["hexDitherFieldPerVisitDec"],
        ) = wrap_ra_dec(sim_data["hexDitherFieldPerVisitRa"], sim_data["hexDitherFieldPerVisitDec"])
        if self.degrees:
            for col in self.cols_added:
                sim_data[col] = np.degrees(sim_data[col])
        return sim_data


class HexDitherFieldPerNightStacker(HexDitherFieldPerVisitStacker):
    """
    Use offsets from the hexagonal grid of 'hexdither', but visit each vertex sequentially.
    Sequential offset for each night of visits.

    Parameters
    ----------
    ra_col : str, optional
        The name of the RA column in the data.
        Default 'fieldRA'.
    dec_col : str, optional
        The name of the Dec column in the data.
        Default 'fieldDec'.
    degrees : bool, optional
        Flag whether RA/Dec should be treated as (and kept as) degrees.
    field_id_col : str, optional
        The name of the fieldId column in the data.
        Used to identify fields which should be identified as the 'same'.
        Default 'fieldId'.
    night_col : str, optional
        The name of the night column in the data.
        Default 'night'.
    max_dither : float, optional
        The radius of the maximum dither offset, in degrees.
        Default 1.75 degrees.
    in_hex : bool, optional
        If True, offsets are constrained to lie within a hexagon inscribed within the max_dither circle.
        If False, offsets can lie anywhere out to the edges of the max_dither circle.
        Default True.
    """

    # Values required for framework operation: this specifies the names of the new columns.
    cols_added = ["hexDitherFieldPerNightRa", "hexDitherFieldPerNightDec"]

    def __init__(
        self,
        ra_col="fieldRA",
        dec_col="fieldDec",
        degrees=True,
        field_id_col="fieldId",
        night_col="night",
        max_dither=1.75,
        in_hex=True,
    ):
        """
        @ MaxDither in degrees
        """
        super().__init__(
            ra_col=ra_col,
            dec_col=dec_col,
            field_id_col=field_id_col,
            degrees=degrees,
            max_dither=max_dither,
            in_hex=in_hex,
        )
        self.night_col = night_col
        # Values required for framework operation: this specifies the data columns required from the database.
        self.cols_req.append(self.night_col)

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            return sim_data
        self._generate_hex_offsets()
        if self.degrees:
            ra = np.radians(sim_data[self.ra_col])
            dec = np.radians(sim_data[self.dec_col])
        else:
            ra = sim_data[self.ra_col]
            dec = sim_data[self.dec_col]
        for fieldid in np.unique(sim_data[self.field_id_col]):
            # Identify observations of this field.
            match = np.where(sim_data[self.field_id_col] == fieldid)[0]
            # Apply a sequential dither, increasing each night.
            vertex_idxs = np.arange(0, len(match), 1)
            nights = sim_data[self.night_col][match]
            vertex_idxs = np.searchsorted(np.unique(nights), nights)
            vertex_idxs = vertex_idxs % self.num_points
            sim_data["hexDitherFieldPerNightRa"][match] = ra[match] + self.x_off[vertex_idxs] / np.cos(
                dec[match]
            )
            sim_data["hexDitherFieldPerNightDec"][match] = dec[match] + self.y_off[vertex_idxs]
        # Wrap into expected range.
        (
            sim_data["hexDitherFieldPerNightRa"],
            sim_data["hexDitherFieldPerNightDec"],
        ) = wrap_ra_dec(sim_data["hexDitherFieldPerNightRa"], sim_data["hexDitherFieldPerNightDec"])
        if self.degrees:
            for col in self.cols_added:
                sim_data[col] = np.degrees(sim_data[col])
        return sim_data


class HexDitherPerNightStacker(HexDitherFieldPerVisitStacker):
    """
    Use offsets from the hexagonal grid of 'hexdither', but visit each vertex sequentially.
    Sequential offset per night for all fields.

    Parameters
    ----------
    ra_col : str, optional
        The name of the RA column in the data.
        Default 'fieldRA'.
    dec_col : str, optional
        The name of the Dec column in the data.
        Default 'fieldDec'.
    degrees : bool, optional
        Flag whether RA/Dec should be treated as (and kept as) degrees.
    field_id_col : str, optional
        The name of the fieldId column in the data.
        Used to identify fields which should be identified as the 'same'.
        Default 'fieldId'.
    night_col : str, optional
        The name of the night column in the data.
        Default 'night'.
    max_dither : float, optional
        The radius of the maximum dither offset, in degrees.
        Default 1.75 degrees.
    in_hex : bool, optional
        If True, offsets are constrained to lie within a hexagon inscribed within the max_dither circle.
        If False, offsets can lie anywhere out to the edges of the max_dither circle.
        Default True.
    """

    # Values required for framework operation: this specifies the names of the new columns.
    cols_added = ["hexDitherPerNightRa", "hexDitherPerNightDec"]

    def __init__(
        self,
        ra_col="fieldRA",
        dec_col="fieldDec",
        degrees=True,
        field_id_col="fieldId",
        night_col="night",
        max_dither=1.75,
        in_hex=True,
    ):
        """
        @ MaxDither in degrees
        """
        super().__init__(
            ra_col=ra_col,
            dec_col=dec_col,
            degrees=degrees,
            field_id_col=field_id_col,
            max_dither=max_dither,
            in_hex=in_hex,
        )
        self.night_col = night_col
        # Values required for framework operation: this specifies the data columns required from the database.
        self.cols_req.append(self.night_col)
        self.added_ra = self.cols_added[0]
        self.added_dec = self.cols_added[1]

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            return sim_data
        # Generate the spiral dither values
        self._generate_hex_offsets()
        nights = np.unique(sim_data[self.night_col])
        if self.degrees:
            ra = np.radians(sim_data[self.ra_col])
            dec = np.radians(sim_data[self.dec_col])
        else:
            ra = sim_data[self.ra_col]
            dec = sim_data[self.dec_col]
        # Add to RA and dec values.
        vertex_id = 0
        for n in nights:
            match = np.where(sim_data[self.night_col] == n)[0]
            vertex_id = vertex_id % self.num_points
            sim_data[self.added_ra][match] = ra[match] + self.x_off[vertex_id] / np.cos(dec[match])
            sim_data[self.added_dec][match] = dec[match] + self.y_off[vertex_id]
            vertex_id += 1
        # Wrap RA/Dec into expected range.
        sim_data[self.added_ra], sim_data[self.added_dec] = wrap_ra_dec(
            sim_data[self.added_ra], sim_data[self.added_dec]
        )
        if self.degrees:
            for col in self.cols_added:
                sim_data[col] = np.degrees(sim_data[col])
        return sim_data


class RandomRotDitherPerFilterChangeStacker(BaseDitherStacker):
    """
    Randomly dither the physical angle of the telescope rotator wrt the mount,
    after every filter change. Visits (in between filter changes) that cannot
    all be assigned an offset without surpassing the rotator limit are not
    dithered.

    Parameters
    ----------
    rot_tel_col : str, optional
        The name of the column in the data specifying the physical angle
        of the telescope rotator wrt. the mount.
        Default: 'rotTelPos'.
    filter_col : str, optional
        The name of the filter column in the data.
        Default: 'filter'.
    degrees : `bool`, optional
        True if angles in the database are in degrees (default).
        If True, returned dithered values are in degrees also.
        If False, angles assumed to be in radians and returned in radians.
    max_dither : float, optional
        Abs(maximum) rotational dither, in degrees. The dithers then will be
        between -max_dither to max_dither.
        Default: 90 degrees.
    max_rot_angle : float, optional
        Maximum rotator angle possible for the camera (degrees). Default 90 degrees.
    min_rot_angle : float, optional
        Minimum rotator angle possible for the camera (degrees). Default -90 degrees.
    random_seed: int, optional
        If set, then used as the random seed for the numpy random number
        generation for the dither offsets.
        Default: None.
    debug: bool, optinal
        If True, will print intermediate steps and plots histograms of
        rotTelPos for cases when no dither is applied.
        Default: False
    """

    # Values required for framework operation: this specifies the names of the new columns.
    cols_added = ["randomDitherPerFilterChangeRotTelPos"]

    def __init__(
        self,
        rot_tel_col="rotTelPos",
        filter_col="filter",
        degrees=True,
        max_dither=90.0,
        max_rot_angle=90,
        min_rot_angle=-90,
        random_seed=None,
        debug=False,
    ):
        # Instantiate the RandomDither object and set internal variables.
        self.rot_tel_col = rot_tel_col
        self.filter_col = filter_col
        self.degrees = degrees
        self.max_dither = max_dither
        self.max_rot_angle = max_rot_angle
        self.min_rot_angle = min_rot_angle
        self.random_seed = random_seed
        # self.units used for plot labels
        if self.degrees:
            self.units = ["deg"]
        else:
            self.units = ["rad"]
            # Convert user-specified values into radians as well.
            self.max_dither = np.radians(self.max_dither)
            self.max_rot_angle = np.radians(self.max_rot_angle)
            self.min_rot_angle = np.radians(self.min_rot_angle)
        self.debug = debug

        # Values required for framework operation: specify the data columns required from the database.
        self.cols_req = [self.rot_tel_col, self.filter_col]

    def _run(self, sim_data, cols_present=False):
        if self.debug:
            import matplotlib.pyplot as plt

        # Just go ahead and return if the columns were already in place.
        if cols_present:
            return sim_data

        # Generate random numbers for dither, using defined seed value if desired.
        # Note that we must define the random state for np.random, to ensure consistency in the build system.
        if not hasattr(self, "_rng"):
            if self.random_seed is not None:
                self._rng = np.random.RandomState(self.random_seed)
            else:
                self._rng = np.random.RandomState(544320)

        if len(np.where(sim_data[self.rot_tel_col] > self.max_rot_angle)[0]) > 0:
            warnings.warn(
                "Input data does not respect the specified maxRotAngle constraint: "
                "(Re)Setting maxRotAngle to max value in the input data: %s" % max(sim_data[self.rot_tel_col])
            )
            self.max_rot_angle = max(sim_data[self.rot_tel_col])
        if len(np.where(sim_data[self.rot_tel_col] < self.min_rot_angle)[0]) > 0:
            warnings.warn(
                "Input data does not respect the specified minRotAngle constraint: "
                "(Re)Setting minRotAngle to min value in the input data: %s" % min(sim_data[self.rot_tel_col])
            )
            self.min_rot_angle = min(sim_data[self.rot_tel_col])

        # Identify points where the filter changes.
        change_idxs = np.where(sim_data[self.filter_col][1:] != sim_data[self.filter_col][:-1])[0]

        # Add the random offsets to the RotTelPos values.
        rot_dither = self.cols_added[0]

        if len(change_idxs) == 0:
            # There are no filter changes, so nothing to dither. Just use original values.
            sim_data[rot_dither] = sim_data[self.rot_tel_col]
        else:
            # For each filter change, generate a series of random values for the offsets,
            # between +/- self.max_dither. These are potential values for the rotational offset.
            # The offset actually used will be  confined to ensure that rotTelPos for all visits in
            # that set of observations (between filter changes) fall within
            # the specified min/maxRotAngle -- without truncating the rotTelPos values.

            # Generate more offsets than needed - either 2x filter changes or 2500, whichever is bigger.
            # 2500 is an arbitrary number.
            max_num = max(len(change_idxs) * 2, 2500)

            rot_offset = np.zeros(len(sim_data), float)
            # Some sets of visits will not be assigned dithers: it was too hard to find an offset.
            n_problematic_ones = 0

            # Loop over the filter change indexes (current filter change, next filter change) to identify
            # sets of visits that should have the same offset.
            for c, cn in zip(change_idxs, change_idxs[1:]):
                random_offsets = self._rng.rand(max_num + 1) * 2.0 * self.max_dither - self.max_dither
                i = 0
                potential_offset = random_offsets[i]
                # Calculate new rotTelPos values, if we used this offset.
                new_rot_tel = sim_data[self.rot_tel_col][c + 1 : cn + 1] + potential_offset
                # Does it work? Do all values fall within minRotAngle / maxRotAngle?
                good_to_go = (new_rot_tel >= self.min_rot_angle).all() and (
                    new_rot_tel <= self.max_rot_angle
                ).all()
                while (not good_to_go) and (i < max_num):
                    # break if find a good offset or hit max_num tries.
                    i += 1
                    potential_offset = random_offsets[i]
                    new_rot_tel = sim_data[self.rot_tel_col][c + 1 : cn + 1] + potential_offset
                    good_to_go = (new_rot_tel >= self.min_rot_angle).all() and (
                        new_rot_tel <= self.max_rot_angle
                    ).all()

                if not good_to_go:  # i.e. no good offset was found after max_num tries
                    n_problematic_ones += 1
                    rot_offset[c + 1 : cn + 1] = 0.0  # no dither
                else:
                    rot_offset[c + 1 : cn + 1] = random_offsets[i]  # assign the chosen offset

            # Handle the last set of observations (after the last filter change to the end of the survey).
            random_offsets = self._rng.rand(max_num + 1) * 2.0 * self.max_dither - self.max_dither
            i = 0
            potential_offset = random_offsets[i]
            new_rot_tel = sim_data[self.rot_tel_col][change_idxs[-1] + 1 :] + potential_offset
            good_to_go = (new_rot_tel >= self.min_rot_angle).all() and (
                new_rot_tel <= self.max_rot_angle
            ).all()
            while (not good_to_go) and (i < max_num):
                # break if find a good offset or cant (after max_num tries)
                i += 1
                potential_offset = random_offsets[i]
                new_rot_tel = sim_data[self.rot_tel_col][change_idxs[-1] + 1 :] + potential_offset
                good_to_go = (new_rot_tel >= self.min_rot_angle).all() and (
                    new_rot_tel <= self.max_rot_angle
                ).all()

            if not good_to_go:  # i.e. no good offset was found after max_num tries
                n_problematic_ones += 1
                rot_offset[c + 1 : cn + 1] = 0.0
            else:
                rot_offset[change_idxs[-1] + 1 :] = potential_offset

        # Assign the dithers
        sim_data[rot_dither] = sim_data[self.rot_tel_col] + rot_offset

        # Final check to make sure things are okay
        good_to_go = (sim_data[rot_dither] >= self.min_rot_angle).all() and (
            sim_data[rot_dither] <= self.max_rot_angle
        ).all()
        if not good_to_go:
            message = "Rotational offsets are not working properly:\n"
            message += " dithered rotTelPos: %s\n" % (sim_data[rot_dither])
            message += " minRotAngle: %s ; maxRotAngle: %s" % (
                self.min_rot_angle,
                self.max_rot_angle,
            )
            raise ValueError(message)
        else:
            return sim_data



__all__ = ("make_bundles_dict_from_list", "MetricBundleGroup")

import os
import sys
import warnings
from collections import OrderedDict

import matplotlib.pyplot as plt
import numpy as np
import numpy.ma as ma
import tqdm

import rubin_sim.maf.db as db
import rubin_sim.maf.utils as utils
from rubin_sim.maf.plots import PlotHandler

from .metric_bundle import MetricBundle, create_empty_metric_bundle


def make_bundles_dict_from_list(bundle_list):
    """Utility to convert a list of MetricBundles into a dictionary, keyed by the fileRoot names.

    Raises an exception if the fileroot duplicates another metricBundle.
    (Note this should alert to potential cases of filename duplication).

    Parameters
    ----------
    bundle_list : `list` of `MetricBundles`
    """
    b_dict = {}
    for b in bundle_list:
        if b.file_root in b_dict:
            raise NameError("More than one metricBundle is using the same fileroot, %s" % (b.file_root))
        b_dict[b.file_root] = b
    return b_dict


class MetricBundleGroup:
    """The MetricBundleGroup exists to calculate the metric values for a group of
    MetricBundles.

    The MetricBundleGroup will query data from a single database table (for multiple
    constraints), use that data to calculate metric values for multiple slicers,
    and calculate summary statistics and generate plots for all metrics included in
    the dictionary passed to the MetricBundleGroup.

    We calculate the metric values here, rather than in the individual MetricBundles,
    because it is much more efficient to step through a slicer once (and calculate all
    the relevant metric values at each point) than it is to repeat this process multiple times.

    The MetricBundleGroup also determines how to efficiently group the MetricBundles
    to reduce the number of sql queries of the database, grabbing larger chunks of data at once.

    Parameters
    ----------
    bundle_dict : `dict` or `list` of `MetricBundles`
        Individual MetricBundles should be placed into a dictionary, and then passed to
        the MetricBundleGroup. The dictionary keys can then be used to identify MetricBundles
        if needed -- and to identify new MetricBundles which could be created if 'reduce'
        functions are run on a particular MetricBundle.
        A bundle_dict can be conveniently created from a list of MetricBundles using
        makeBundlesDictFromList (done automatically if a list is passed in)
    db_con : `str` or database connection object
        A str that is the path to a sqlite3 file or a database object that can be used by pandas.read_sql.
        Advanced use: It is possible to set this to None, in which case data should be passed
        directly to the runCurrent method (and runAll should not be used).
    out_dir : `str`, optional
        Directory to save the metric results. Default is the current directory.
    results_db : `ResultsDb`, optional
        A results database. If not specified, one will be created in the out_dir.
        This database saves information about the metrics calculated, including their summary statistics.
    verbose : `bool`, optional
        Flag to turn on/off verbose feedback.
    save_early : `bool`, optional
        If True, metric values will be saved immediately after they are first calculated (to prevent
        data loss) as well as after summary statistics are calculated.
        If False, metric values will only be saved after summary statistics are calculated.
    db_table : `str`, optional
        The name of the table in the db_obj to query for data.
    """

    def __init__(
        self,
        bundle_dict,
        db_con,
        out_dir=".",
        results_db=None,
        verbose=False,
        save_early=True,
        db_table=None,
    ):
        """Set up the MetricBundleGroup."""
        if type(bundle_dict) is list:
            bundle_dict = make_bundles_dict_from_list(bundle_dict)
        # Print occasional messages to screen.
        self.verbose = verbose
        # Save metric results as soon as possible (in case of crash).
        self.save_early = save_early
        # Check for output directory, create it if needed.
        self.out_dir = out_dir
        if not os.path.isdir(self.out_dir):
            os.makedirs(self.out_dir)

        # Do some type checking on the MetricBundle dictionary.
        if not isinstance(bundle_dict, dict):
            raise ValueError("bundleDict should be a dictionary containing MetricBundle objects.")
        for b in bundle_dict.values():
            if not isinstance(b, MetricBundle):
                raise ValueError("bundleDict should contain only MetricBundle objects.")
        # Identify the series of constraints.
        self.constraints = list(set([b.constraint for b in bundle_dict.values()]))
        # Set the bundleDict (all bundles, with all constraints)
        self.bundle_dict = bundle_dict

        self.db_obj = db_con
        # Set the table we're going to be querying.
        self.db_table = db_table

        # Check the results_db (optional).
        if results_db is not None:
            if not isinstance(results_db, db.ResultsDb):
                raise ValueError("results_db should be an ResultsDb object")
        self.results_db = results_db

        # Dict to keep track of what's been run:
        self.has_run = {}
        for bk in bundle_dict:
            self.has_run[bk] = False

    def _check_compatible(self, metric_bundle1, metric_bundle2):
        """Check if two MetricBundles are "compatible".
        Compatible indicates that the sql constraints, the slicers, and the maps are the same, and
        that the stackers do not interfere with each other
        (i.e. are not trying to set the same column in different ways).
        Returns True if the MetricBundles are compatible, False if not.

        Parameters
        ----------
        metric_bundle1 : MetricBundle
        metric_bundle2 : MetricBundle

        Returns
        -------
        bool
        """
        if metric_bundle1.constraint != metric_bundle2.constraint:
            return False
        if metric_bundle1.slicer != metric_bundle2.slicer:
            return False
        if metric_bundle1.maps_list.sort() != metric_bundle2.maps_list.sort():
            return False
        for stacker in metric_bundle1.stacker_list:
            for stacker2 in metric_bundle2.stacker_list:
                # If the stackers have different names, that's OK, and if they are identical, that's ok.
                if (stacker.__class__.__name__ == stacker2.__class__.__name__) & (stacker != stacker2):
                    return False
        # But if we got this far, everything matches.
        return True

    def _find_compatible_lists(self):
        """Find sets of compatible metricBundles from the currentBundleDict."""
        # CompatibleLists stores a list of lists;
        #   each (nested) list contains the bundleDict _keys_ of a compatible set of metricBundles.
        #
        compatible_lists = []
        for k, b in self.current_bundle_dict.items():
            found_compatible = False
            for compatible_list in compatible_lists:
                comparison_metric_bundle_key = compatible_list[0]
                compatible = self._check_compatible(self.bundle_dict[comparison_metric_bundle_key], b)
                if compatible:
                    # Must compare all metricBundles in each subset (if they are a potential match),
                    #  as the stackers could be different (and one could be incompatible,
                    #  not necessarily the first)
                    for comparison_metric_bundle_key in compatible_list[1:]:
                        compatible = self._check_compatible(self.bundle_dict[comparison_metric_bundle_key], b)
                        if not compatible:
                            # If we find one which is not compatible, stop and go on to the
                            # next subset list.
                            break
                    # Otherwise, we reached the end of the subset and they were all compatible.
                    found_compatible = True
                    compatible_list.append(k)
            if not found_compatible:
                # Didn't find a pre-existing compatible set; make a new one.
                compatible_lists.append(
                    [
                        k,
                    ]
                )
        self.compatible_lists = compatible_lists

    def run_all(self, clear_memory=False, plot_now=False, plot_kwargs=None):
        """Runs all the metricBundles in the metricBundleGroup, over all constraints.

        Calculates metric values, then runs reduce functions and summary statistics for
        all MetricBundles.

        Parameters
        ----------
        clear_memory : `bool`, optional
            If True, deletes metric values from memory after running each constraint group.
        plot_now : `bool`, optional
            If True, plots the metric values immediately after calculation.
        plot_kwargs : `bool`, optional
            kwargs to pass to plotCurrent.
        """
        for constraint in self.constraints:
            # Set the 'currentBundleDict' which is a dictionary of the metricBundles which match this
            #  constraint.
            self.run_current(
                constraint,
                clear_memory=clear_memory,
                plot_now=plot_now,
                plot_kwargs=plot_kwargs,
            )

    def set_current(self, constraint):
        """Utility to set the currentBundleDict (i.e. a set of metricBundles with the same SQL constraint).

        Parameters
        ----------
        constraint : `str`
            The subset of MetricBundles with metricBundle.constraint == constraint will be
            included in a subset identified as the currentBundleDict.
            These are the active metrics to be calculated and plotted, etc.
        """
        if constraint is None:
            constraint = ""
        self.current_bundle_dict = {}
        for k, b in self.bundle_dict.items():
            if b.constraint == constraint:
                self.current_bundle_dict[k] = b
        # Build list of all the columns needed from the database.
        self.db_cols = []
        for b in self.current_bundle_dict.values():
            self.db_cols.extend(b.db_cols)
        self.db_cols = list(set(self.db_cols))

    def run_current(
        self,
        constraint,
        sim_data=None,
        clear_memory=False,
        plot_now=False,
        plot_kwargs=None,
    ):
        """Run all the metricBundles which match this constraint in the
        metricBundleGroup.

        Calculates the metric values, then runs reduce functions and
        summary statistics for metrics in the current set only
        (see self.setCurrent).

        Parameters
        ----------
        constraint : `str`
           constraint to use to set the currently active metrics
        sim_data : `numpy.ndarray`, optional
           If simData is not None, then this numpy structured array is used
           instead of querying data from the dbObj.
        clear_memory : `bool`, optional
           If True, metric values are deleted from memory after they are
           calculated (and saved to disk).
        plot_now : `bool`, optional
           Plot immediately after calculating metric values
           (instead of the usual procedure, which is to plot after metric
           values are calculated for all constraints).
        plot_kwargs : kwargs, optional
           Plotting kwargs to pass to plotCurrent.
        """
        self.set_current(constraint)

        # Can pass simData directly (if had other method for getting data)
        if sim_data is not None:
            self.sim_data = sim_data

        else:
            self.sim_data = None
            # Query for the data.
            try:
                self.get_data(constraint)
            except UserWarning:
                warnings.warn("No data matching constraint %s" % constraint)
                metrics_skipped = []
                for b in self.current_bundle_dict.values():
                    metrics_skipped.append(
                        "%s : %s : %s" % (b.metric.name, b.info_label, b.slicer.slicer_name)
                    )
                warnings.warn(" This means skipping metrics %s" % metrics_skipped)
                return
            except ValueError:
                warnings.warn(
                    "One or more of the columns requested from the database was not available."
                    + " Skipping constraint %s" % constraint
                )
                metrics_skipped = []
                for b in self.current_bundle_dict.values():
                    metrics_skipped.append(
                        "%s : %s : %s" % (b.metric.name, b.info_label, b.slicer.slicer_name)
                    )
                warnings.warn(" This means skipping metrics %s" % metrics_skipped)
                return

        # Find compatible subsets of the MetricBundle dictionary,
        # which can be run/metrics calculated/ together.
        self._find_compatible_lists()

        for compatible_list in self.compatible_lists:
            if self.verbose:
                print("Running: ", compatible_list)
            self._run_compatible(compatible_list)
            if self.verbose:
                print("Completed metric generation.")
            for key in compatible_list:
                self.has_run[key] = True
        # Run the reduce methods.
        if self.verbose:
            print("Running reduce methods.")
        self.reduce_current()
        # Run the summary statistics.
        if self.verbose:
            print("Running summary statistics.")
        self.summary_current()
        if self.verbose:
            print("Completed.")
        if plot_now:
            if plot_kwargs is None:
                self.plot_current()
            else:
                self.plot_current(**plot_kwargs)
        # Optionally: clear results from memory.
        if clear_memory:
            for b in self.current_bundle_dict.values():
                b.metric_values = None
            if self.verbose:
                print("Deleted metric_values from memory.")

    def get_data(self, constraint):
        """Query the data from the database.

        The currently bundleDict should generally be set
        before calling getData (using setCurrent).

        Parameters
        ----------
        constraint : `str`
           The constraint for the currently active set of MetricBundles.
        """
        if self.verbose:
            if constraint == "":
                print("Querying table %s with no constraint for columns %s." % (self.db_table, self.db_cols))
            else:
                print(
                    "Querying table %s with constraint %s for columns %s"
                    % (self.db_table, constraint, self.db_cols)
                )
        # Note that we do NOT run the stackers at this point
        # (this must be done in each 'compatible' group).
        self.sim_data = utils.get_sim_data(
            self.db_obj,
            constraint,
            self.db_cols,
            table_name=self.db_table,
        )

        if self.verbose:
            print("Found %i visits" % (self.sim_data.size))

    def _run_compatible(self, compatible_list):
        """Runs a set of 'compatible' metricbundles in the MetricBundleGroup
        dictionary identified by 'compatible_list' keys.

        A compatible list of MetricBundles is a subset of the currentBundleDict.
        The currentBundleDict == set of MetricBundles with the same constraint.
        The compatibleBundles == set of MetricBundles with the same constraint,
        the same slicer, the same maps applied to the slicer,
        and stackers which do not clobber each other's data.

        This is where the work of calculating the metric values is done.
        """

        if len(self.sim_data) == 0:
            return

        # Grab a dictionary representation of this subset of the dictionary,
        # for easier iteration.
        b_dict = {key: self.current_bundle_dict.get(key) for key in compatible_list}

        # Find the unique stackers and maps.
        # These are already "compatible" (as id'd by compatible_list).
        uniq_stackers = []
        all_stackers = []
        uniq_maps = []
        all_maps = []
        for b in b_dict.values():
            all_stackers += b.stacker_list
            all_maps += b.maps_list
        for s in all_stackers:
            if s not in uniq_stackers:
                uniq_stackers.append(s)
        for m in all_maps:
            if m not in uniq_maps:
                uniq_maps.append(m)

        # Run stackers.
        for stacker in uniq_stackers:
            # Note that stackers will clobber previously existing columns
            self.sim_data = stacker.run(self.sim_data, override=True)

        # Pull out one of the slicers to use as our 'slicer'.
        # This will be forced back into all of the metricBundles
        # at the end (so that they track the same info_label such as the
        # slice_points, in case the same actual object wasn't used).
        slicer = list(b_dict.values())[0].slicer
        slicer.setup_slicer(self.sim_data, maps=uniq_maps)
        # Copy the slicer (after setup) back into the individual metricBundles.
        if slicer.slicer_name != "HealpixSlicer" or slicer.slicer_name != "UniSlicer":
            for b in b_dict.values():
                b.slicer = slicer

        # Set up (masked) arrays to store metric data in each metricBundle.
        for b in b_dict.values():
            b._setup_metric_values()

        # Set up an ordered dictionary to be the cache if needed:
        # (Currently using OrderedDict,
        # it might be faster to use 2 regular Dicts instead)
        if slicer.cache_size > 0:
            cache_dict = OrderedDict()
            cache = True
        else:
            cache = False
        # Run through all slicepoints and calculate metrics.
        if self.verbose:
            slicer_iter = tqdm.tqdm(
                slicer,
                desc="Processing slices",
                ncols=79,
                file=sys.stdout,
            )
        else:
            slicer_iter = slicer

        for slice_i in slicer_iter:
            i = slice_i["slice_point"]["sid"]
            slicedata = self.sim_data[slice_i["idxs"]]
            if len(slicedata) == 0:
                # No data at this slice_point. Mask data values.
                for b in b_dict.values():
                    b.metric_values.mask[i] = True
            else:
                # There is data! Should we use our data cache?
                if cache:
                    # Make the data idxs hashable.
                    cache_key = frozenset(slice_i["idxs"])
                    # If key exists, set flag to use it, otherwise add it
                    if cache_key in cache_dict:
                        use_cache = True
                        cache_val = cache_dict[cache_key]
                        # Move this value to the end of the OrderedDict
                        del cache_dict[cache_key]
                        cache_dict[cache_key] = cache_val
                    else:
                        cache_dict[cache_key] = i
                        use_cache = False
                    for b in b_dict.values():
                        if use_cache:
                            b.metric_values.data[i] = b.metric_values.data[cache_dict[cache_key]]
                        else:
                            b.metric_values.data[i] = b.metric.run(
                                slicedata, slice_point=slice_i["slice_point"]
                            )
                    # If we are above the cache size,
                    # drop the oldest element from the cache dict.
                    if len(cache_dict) > slicer.cache_size:
                        del cache_dict[list(cache_dict.keys())[0]]

                # Not using memoize, just calculate things normally
                else:
                    for b in b_dict.values():
                        try:
                            b.metric_values.data[i] = b.metric.run(
                                slicedata, slice_point=slice_i["slice_point"]
                            )
                        except BaseException as e:
                            print(f"Failed at slice_point {slice_i}, sid {i}")
                            raise e
        # Mask data where metrics could not be computed
        # (according to metric bad value).
        for b in b_dict.values():
            if b.metric_values.dtype.name == "object":
                for ind, val in enumerate(b.metric_values.data):
                    if val is b.metric.badval:
                        b.metric_values.mask[ind] = True
            else:
                # For some reason, this doesn't work for dtype=object arrays.
                b.metric_values.mask = np.where(
                    b.metric_values.data == b.metric.badval, True, b.metric_values.mask
                )

        # Save data to disk as we go (just failsafe).
        if self.save_early:
            for b in b_dict.values():
                b.write(out_dir=self.out_dir, results_db=self.results_db)
        else:
            # Just write the metric run information to the results_db
            for b in b_dict.values():
                b.write_db(results_db=self.results_db)

    def reduce_all(self, update_summaries=True):
        """Run the reduce methods for all metrics in bundleDict.

        Running this method, for all MetricBundles at once,
        assumes that clearMemory was False.

        Parameters
        ----------
        update_summaries : `bool`, optional
            If True, summary metrics are removed from the top-level
            (non-reduced) MetricBundle. Usually this should be True,
            as summary metrics are generally intended to run on the simpler
            data produced by reduce metrics.
        """
        for constraint in self.constraints:
            self.set_current(constraint)
            self.reduce_current(update_summaries=update_summaries)

    def reduce_current(self, update_summaries=True):
        """Run all reduce functions for the metricbundle in the
        currently active set of MetricBundles.

        Parameters
        ----------
        update_summaries : `bool`, optional
            If True, summary metrics are removed from the top-level
            (non-reduced) MetricBundle. Usually this should be True,
            as summary metrics are generally intended to run on the simpler
            data produced by reduce metrics.
        """
        # Create a temporary dictionary to hold the reduced metricbundles.
        reduce_bundle_dict = {}
        for b in self.current_bundle_dict.values():
            # If there are no reduce functions associated with the metric, skip this metricBundle.
            if len(b.metric.reduce_funcs) > 0:
                # Apply reduce functions, creating a new metricBundle in the process (new metric values).
                for r in b.metric.reduce_funcs:
                    newmetricbundle = b.reduce_metric(b.metric.reduce_funcs[r], reduce_func_name=r)
                    # Add the new metricBundle to our metricBundleGroup dictionary.
                    name = newmetricbundle.metric.name
                    if name in self.bundle_dict:
                        name = newmetricbundle.file_root
                    reduce_bundle_dict[name] = newmetricbundle
                    if self.save_early:
                        newmetricbundle.write(out_dir=self.out_dir, results_db=self.results_db)
                    else:
                        newmetricbundle.write_db(results_db=self.results_db)
                # Remove summaryMetrics from top level metricbundle if desired.
                if update_summaries:
                    b.summary_metrics = []
        # Add the new metricBundles to the MetricBundleGroup dictionary.
        self.bundle_dict.update(reduce_bundle_dict)
        # And add to to the currentBundleDict too, so we run as part of 'summaryCurrent'.
        self.current_bundle_dict.update(reduce_bundle_dict)

    def summary_all(self):
        """Run the summary statistics for all metrics in bundleDict.

        Calculating all summary statistics, for all MetricBundles, at this
        point assumes that clearMemory was False.
        """
        for constraint in self.constraints:
            self.set_current(constraint)
            self.summary_current()

    def summary_current(self):
        """Run summary statistics on all the metricBundles in the
        currently active set of MetricBundles."""
        for b in self.current_bundle_dict.values():
            b.compute_summary_stats(self.results_db)

    def plot_all(
        self,
        save_figs=True,
        outfile_suffix=None,
        fig_format="pdf",
        dpi=600,
        trim_whitespace=True,
        thumbnail=True,
        closefigs=True,
    ):
        """Generate all the plots for all the metricBundles in bundleDict.

        Generating all plots, for all MetricBundles,
        At this point, assumes that clearMemory was False.

        Parameters
        ----------
        savefig : `bool`, optional
            If True, save figures to disk, to self.out_dir directory.
        outfile_suffix : `str`, optional
            Append outfile_suffix to the end of every plot file generated.
            Useful for generating sequential series of images for movies.
        fig_format : `str`, optional
            Matplotlib figure format to use to save to disk.
        dpi : `int`, optional
            DPI for matplotlib figure.
        trim_whitespace : `bool`, optional
            If True, trim additional whitespace from final figures.
        thumbnail : `bool`, optional
            If True, save a small thumbnail jpg version of the output file
            to disk as well. This is useful for show_maf web pages.
        closefigs : `bool`, optional
            Close the matplotlib figures after they are saved to disk.
            If many figures are generated, closing the figures saves
            significant memory.
        """
        for constraint in self.constraints:
            if self.verbose:
                print('Plotting figures with "%s" constraint now.' % (constraint))

            self.set_current(constraint)
            self.plot_current(
                savefig=save_figs,
                outfile_suffix=outfile_suffix,
                fig_format=fig_format,
                dpi=dpi,
                trim_whitespace=trim_whitespace,
                thumbnail=thumbnail,
                closefigs=closefigs,
            )

    def plot_current(
        self,
        savefig=True,
        outfile_suffix=None,
        fig_format="pdf",
        dpi=600,
        trim_whitespace=True,
        thumbnail=True,
        closefigs=True,
    ):
        """Generate the plots for the currently active set of MetricBundles.

        Parameters
        ----------
        savefig : `bool`, optional
            If True, save figures to disk, to self.out_dir directory.
        outfile_suffix : `str`, optional
            Append outfile_suffix to the end of every plot file generated.
            Useful for generating sequential series of images for movies.
        fig_format : `str`, optional
            Matplotlib figure format to use to save to disk.
        dpi : `int`, optional
            DPI for matplotlib figure.
        trim_whitespace : `bool`, optional
            If True, trim additional whitespace from final figures.
        thumbnail : `bool`, optional
            If True, save a small thumbnail jpg version of the output file
            to disk as well. This is useful for show_maf web pages.
        closefigs : `bool`, optional
            Close the matplotlib figures after they are saved to disk.
            If many figures are generated, closing the figures saves
            significant memory.
        """
        plot_handler = PlotHandler(
            out_dir=self.out_dir,
            results_db=self.results_db,
            savefig=savefig,
            fig_format=fig_format,
            dpi=dpi,
            trim_whitespace=trim_whitespace,
            thumbnail=thumbnail,
        )

        for b in self.current_bundle_dict.values():
            try:
                b.plot(
                    plot_handler=plot_handler,
                    outfile_suffix=outfile_suffix,
                    savefig=savefig,
                )
            except ValueError as ve:
                message = "Plotting failed for metricBundle %s." % (b.file_root)
                message += " Error message: %s" % (ve)
                warnings.warn(message)
            if closefigs:
                plt.close("all")
        if self.verbose:
            print("Plotting complete.")

    def write_all(self):
        """Save all the MetricBundles to disk.

        Saving all MetricBundles to disk at this point assumes that clearMemory was False.
        """
        for constraint in self.constraints:
            self.set_current(constraint)
            self.write_current()

    def write_current(self):
        """Save all the MetricBundles in the currently active set to disk."""
        if self.verbose:
            if self.save_early:
                print("Re-saving metric bundles.")
            else:
                print("Saving metric bundles.")
        for b in self.current_bundle_dict.values():
            b.write(out_dir=self.out_dir, results_db=self.results_db)

    def read_all(self):
        """Attempt to read all MetricBundles from disk.

        You must set the metrics/slicer/constraint/run_name for a metricBundle
        appropriately, so that the file_root is correct.
        """
        reduce_bundle_dict = {}
        remove_bundles = []
        for b in self.bundle_dict:
            bundle = self.bundle_dict[b]
            filename = os.path.join(self.out_dir, bundle.file_root + ".npz")
            try:
                # Create a temporary metricBundle to read the data into.
                # We are not using "b" directly, as we are trying NOT
                # to override the plot_dict or display_dict.
                tmp_bundle = create_empty_metric_bundle()
                tmp_bundle.read(filename)
                # Copy the tmp_bundle metric_values into bundle.
                bundle.metric_values = tmp_bundle.metric_values
                # And copy the slicer into b, to get slice_points.
                bundle.slicer = tmp_bundle.slicer
                # Copy the summary stats though.
                bundle.summary_values = tmp_bundle.summary_values
                if self.verbose:
                    print("Read %s from disk." % (bundle.file_root))
            except IOError:
                warnings.warn("Warning: file %s not found, bundle not restored." % filename)
                remove_bundles.append(b)

            # Look to see if this is a complex metric,
            # with associated 'reduce' functions, and read those in too.
            if len(bundle.metric.reduce_funcs) > 0:
                orig_metric_name = bundle.metric.name
                for reduce_func in bundle.metric.reduce_funcs.values():
                    reduce_name = orig_metric_name + "_" + reduce_func.__name__.replace("reduce", "")
                    # Borrow the fileRoot in b (we'll reset it appropriately)
                    bundle.metric.name = reduce_name
                    bundle._build_file_root()
                    filename = os.path.join(self.out_dir, bundle.file_root + ".npz")
                    tmp_bundle = create_empty_metric_bundle()
                    try:
                        tmp_bundle.read(filename)
                        # This won't necessarily recreate the plot_dict and
                        # display_dict exactly as they would have been made
                        # if you calculated the reduce metric from scratch.
                        newmetric_bundle = MetricBundle(
                            metric=bundle.metric,
                            slicer=bundle.slicer,
                            constraint=bundle.constraint,
                            stacker_list=bundle.stacker_list,
                            run_name=bundle.run_name,
                            info_label=bundle.info_label,
                            plot_dict=bundle.plotDict,
                            display_dict=bundle.displayDict,
                            summary_metrics=bundle.summary_metrics,
                            maps_list=bundle.maps_list,
                            file_root=bundle.file_root,
                            plot_funcs=bundle.plot_funcs,
                        )
                        newmetric_bundle.metric.name = reduce_name
                        newmetric_bundle.metric_values = ma.copy(tmp_bundle.metric_values)
                        newmetric_bundle.summary_values = tmp_bundle.summary_values
                        # Add the new metricBundle to metricBundleGroup dict.
                        name = newmetric_bundle.metric.name
                        if name in self.bundle_dict:
                            name = newmetric_bundle.file_root
                        reduce_bundle_dict[name] = newmetric_bundle
                        if self.verbose:
                            print("Read %s from disk." % (newmetric_bundle.file_root))
                    except IOError:
                        warnings.warn(
                            'Warning: file %s not found, bundle not restored ("reduce" metric).' % filename
                        )

                    # Remove summaryMetrics from top level metricbundle.
                    bundle.summary_metrics = []
                    # Update parent MetricBundle name.
                    bundle.metric.name = orig_metric_name
                    bundle._build_file_root()

        # Add the reduce bundles into the bundleDict.
        self.bundle_dict.update(reduce_bundle_dict)
        # And remove the bundles which were not found on disk,
        # so we don't try to make (blank) plots.
        for b in remove_bundles:
            del self.bundle_dict[b]



__all__ = ("run_moving_join",)

import argparse
import glob
import os

import matplotlib
import numpy as np

matplotlib.use("Agg")

from . import batches as batches


def run_moving_join():
    """Join split metric outputs into a single metric output file."""
    parser = argparse.ArgumentParser(
        description="Join moving object metrics (from splits) for a particular "
        "scheduler run.  Assumes split metric files are in "
        "<orbitRoot_split> subdirectories of base_dir. "
    )
    parser.add_argument("--orbit_file", type=str, help="File containing the moving object orbits.")
    parser.add_argument(
        "--base_dir",
        type=str,
        default=".",
        help="Root directory containing split (or single) metric outputs.",
    )
    parser.add_argument(
        "--out_dir",
        type=str,
        default=None,
        help="Output directory for moving object metrics. Default [orbitRoot]",
    )
    args = parser.parse_args()

    if args.orbit_file is None:
        print("Must specify an orbit_file")
        exit()

    # Outputs from the metrics are generally like so:
    #  <base_dir>/<splitDir>/<metricFileName>
    # - base_dir tends to be <opsimName_orbitRoot> (but is set by user when starting to generate obs.)
    # - splitDir tends to be <orbitRoot_split#> (and is set by observation generation script)
    # - metricFile is <opsimName_metricName_metadata(NEO/L7/etc + metadata from metric script)_MOOB.npz
    #  (the metricFileName is set by the metric generation script - run_moving_calc.py).
    #  (note that split# does not show up in the metricFileName, and is not used in run_moving_calc.py).
    #  ... this lets run_moving_calc.py easily run in parallel on multiple splits.

    # Assume splits = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    splits = np.arange(0, 10, 1)
    orbit_root = args.orbit_file.replace(".txt", "").replace(".des", "").replace(".s3m", "")

    if args.out_dir is not None:
        out_dir = args.out_dir
    else:
        out_dir = f"{orbit_root}"

    # Scan first splitDir for all metric files.
    tempdir = os.path.join(args.base_dir, f"{orbit_root}_{splits[0]}")
    print(f"# Joining files from {orbit_root}_[0-9]; will use {tempdir} to find metric names.")

    metricfiles = glob.glob(os.path.join(tempdir, "*MOOB.npz"))
    # Identify metric names that we want to join.
    metricNames = []
    for m in metricfiles:
        mname = os.path.split(m)[-1]
        # Hack out raw Discovery outputs. We don't want to join the raw discovery files.
        # This is a hack because currently we're just pulling out _Time and _N_Chances to join.
        if "Discovery" in mname:
            if "DiscoveryTime" in mname:
                metricNames.append(mname)
            elif "DiscoveryNChances" in mname:
                metricNames.append(mname)
            elif "Magic" in mname:
                metricNames.append(mname)
            elif "HighVelocity" in mname:
                metricNames.append(mname)
            else:
                pass
        else:
            metricNames.append(mname)

    if len(metricNames) == 0:
        print(f"Could not read any metric files from {tempdir}")
        exit()

    # Create the output directory.
    if not (os.path.isdir(out_dir)):
        os.makedirs(out_dir)

    # Read and combine the metric files.
    for m in metricNames:
        b = batches.read_and_combine(orbit_root, args.base_dir, splits, m)
        b.write(out_dir=out_dir)



__all__ = (
    "ParallaxMetric",
    "ProperMotionMetric",
    "RadiusObsMetric",
    "ParallaxCoverageMetric",
    "ParallaxDcrDegenMetric",
)

import numpy as np
import rubin_scheduler.utils as utils
from scipy.optimize import curve_fit

import rubin_sim.maf.utils as mafUtils

from .base_metric import BaseMetric


class ParallaxMetric(BaseMetric):
    """Calculate the uncertainty in a parallax measurement
    given a series of observations.

    Uses columns ra_pi_amp and dec_pi_amp,
    calculated by the ParallaxFactorStacker.

    Parameters
    ----------
    m5_col : `str`, optional
        The default column name for m5 information in the input data.
    filter_col : `str`, optional
        The column name for the filter information.
    seeing_col : `str`, optional
        The column name for the seeing information.
        Since the astrometry errors are based on the physical size of the PSF,
        this should be the FWHM of the physical psf, e.g. seeingFwhmGeom.
    rmag : `float`, optional
        The r magnitude of the fiducial star in r band.
        Other filters are scaled using sedTemplate keyword.
    SedTemplate : `str`, optional
        The template to use. This can be 'flat' or 'O','B','A','F','G','K','M'.
    atm_err : `float`, optional
        The expected centroiding error due to the atmosphere, in arcseconds.
        Default 0.01.
    normalize : `bool`, optional
        Compare the astrometric uncertainty to the uncertainty
        that would result if half the observations were taken at the start
        and half at the end.
        A perfect survey will have a value close to 1, while
        a poorly scheduled survey will be close to 0.
    badval : `float`, optional
        The value to return when the metric value cannot be calculated.
    """

    def __init__(
        self,
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        seeing_col="seeingFwhmGeom",
        rmag=20.0,
        sed_template="flat",
        badval=-666,
        atm_err=0.01,
        normalize=False,
        **kwargs,
    ):
        cols = [m5_col, filter_col, seeing_col, "ra_pi_amp", "dec_pi_amp"]
        if normalize:
            units = "ratio"
        else:
            units = "mas"
        super().__init__(cols, units=units, badval=badval, **kwargs)
        # set return type
        self.m5_col = m5_col
        self.seeing_col = seeing_col
        self.filter_col = filter_col
        filters = ["u", "g", "r", "i", "z", "y"]
        self.mags = {}
        if sed_template == "flat":
            for f in filters:
                self.mags[f] = rmag
        else:
            self.mags = utils.stellarMags(sed_template, rmag=rmag)
        self.atm_err = atm_err
        self.normalize = normalize
        self.comment = (
            "Estimated uncertainty in parallax measurement "
            "(assuming no proper motion or that proper motion "
        )
        self.comment += (
            "is well fit). Uses measurements in all bandpasses, "
            "and estimates astrometric error based on SNR "
        )
        self.comment += "in each visit. "
        if sed_template == "flat":
            self.comment += "Assumes a flat SED. "
        if self.normalize:
            self.comment += (
                "This normalized version of the metric displays the "
                "estimated uncertainty in the parallax measurement, "
            )
            self.comment += "divided by the minimum parallax uncertainty possible " "(if all visits were six "
            self.comment += (
                "months apart). Values closer to 1 indicate more optimal "
                "scheduling for parallax measurement."
            )

    def _final_sigma(self, position_errors, ra_pi_amp, dec_pi_amp):
        """Assume parallax in RA and DEC are fit independently, then combined.
        All inputs assumed to be arcsec"""
        with np.errstate(divide="ignore", invalid="ignore"):
            sigma_a = position_errors / ra_pi_amp
            sigma_b = position_errors / dec_pi_amp
            sigma_ra = np.sqrt(1.0 / np.sum(1.0 / sigma_a**2))
            sigma_dec = np.sqrt(1.0 / np.sum(1.0 / sigma_b**2))
            # Combine RA and Dec uncertainties, convert to mas
            sigma = np.sqrt(1.0 / (1.0 / sigma_ra**2 + 1.0 / sigma_dec**2)) * 1e3
        return sigma

    def run(self, data_slice, slice_point=None):
        filters = np.unique(data_slice[self.filter_col])
        if hasattr(filters[0], "decode"):
            filters = [str(f.decode("utf-8")) for f in filters]
        snr = np.zeros(len(data_slice), dtype="float")
        # compute SNR for all observations
        for filt in filters:
            good = np.where(data_slice[self.filter_col] == filt)
            snr[good] = mafUtils.m52snr(self.mags[str(filt)], data_slice[self.m5_col][good])
        position_errors = mafUtils.astrom_precision(data_slice[self.seeing_col], snr, self.atm_err)
        sigma = self._final_sigma(position_errors, data_slice["ra_pi_amp"], data_slice["dec_pi_amp"])
        if self.normalize:
            # Leave the dec parallax as zero since one can't have ra and dec maximized at the same time.
            sigma = (
                self._final_sigma(
                    position_errors,
                    data_slice["ra_pi_amp"] * 0 + 1.0,
                    data_slice["dec_pi_amp"] * 0,
                )
                / sigma
            )
        return sigma


class ProperMotionMetric(BaseMetric):
    """Calculate the uncertainty in the returned proper motion.

    This metric assumes gaussian errors in the astrometry measurements.

    Parameters
    ----------
    metricName : str, optional
        Default 'properMotion'.
    m5_col : str, optional
        The default column name for m5 information in the input data. Default fiveSigmaDepth.
    mjd_col : str, optional
        The column name for the exposure time. Default observationStartMJD.
    filterCol : str, optional
        The column name for the filter information. Default filter.
    seeing_col : str, optional
        The column name for the seeing information. Since the astrometry errors are based on the physical
        size of the PSF, this should be the FWHM of the physical psf. Default seeingFwhmGeom.
    rmag : float, optional
        The r magnitude of the fiducial star in r band. Other filters are sclaed using sedTemplate keyword.
        Default 20.0
    SedTemplate : str, optional
        The template to use. This can be 'flat' or 'O','B','A','F','G','K','M'. Default flat.
    atm_err : float, optional
        The expected centroiding error due to the atmosphere, in arcseconds. Default 0.01.
    normalize : `bool`, optional
        Compare the astrometric uncertainty to the uncertainty that would result if half the observations
        were taken at the start and half at the end. A perfect survey will have a value close to 1, while
        a poorly scheduled survey will be close to 0. Default False.
    baseline : float, optional
        The length of the survey used for the normalization, in years. Default 10.
    badval : float, optional
        The value to return when the metric value cannot be calculated. Default -666.
    """

    def __init__(
        self,
        metric_name="properMotion",
        m5_col="fiveSigmaDepth",
        mjd_col="observationStartMJD",
        filter_col="filter",
        seeing_col="seeingFwhmGeom",
        rmag=20.0,
        sed_template="flat",
        badval=-666,
        atm_err=0.01,
        normalize=False,
        baseline=10.0,
        **kwargs,
    ):
        cols = [m5_col, mjd_col, filter_col, seeing_col]
        if normalize:
            units = "ratio"
        else:
            units = "mas/yr"
        super(ProperMotionMetric, self).__init__(
            col=cols, metric_name=metric_name, units=units, badval=badval, **kwargs
        )
        # set return type
        self.mjd_col = mjd_col
        self.seeing_col = seeing_col
        self.m5_col = m5_col
        filters = ["u", "g", "r", "i", "z", "y"]
        self.mags = {}
        if sed_template == "flat":
            for f in filters:
                self.mags[f] = rmag
        else:
            self.mags = utils.stellarMags(sed_template, rmag=rmag)
        self.atm_err = atm_err
        self.normalize = normalize
        self.baseline = baseline
        self.comment = (
            "Estimated uncertainty of the proper motion fit "
            "(assuming no parallax or that parallax is well fit). "
        )
        self.comment += (
            "Uses visits in all bands, and generates approximate "
            "astrometric errors using the SNR in each visit. "
        )
        if sed_template == "flat":
            self.comment += "Assumes a flat SED. "
        if self.normalize:
            self.comment += (
                "This normalized version of the metric represents " "the estimated uncertainty in the proper "
            )
            self.comment += "motion divided by the minimum uncertainty possible " "(if all visits were "
            self.comment += "obtained on the first and last days of the survey). "
            self.comment += "Values closer to 1 indicate more optimal scheduling."

    def run(self, data_slice, slice_point=None):
        filters = np.unique(data_slice["filter"])
        filters = [str(f) for f in filters]
        precis = np.zeros(data_slice.size, dtype="float")
        for f in filters:
            observations = np.where(data_slice["filter"] == f)
            if np.size(observations[0]) < 2:
                precis[observations] = self.badval
            else:
                snr = mafUtils.m52snr(self.mags[f], data_slice[self.m5_col][observations])
                precis[observations] = mafUtils.astrom_precision(
                    data_slice[self.seeing_col][observations], snr, self.atm_err
                )
        good = np.where(precis != self.badval)
        result = mafUtils.sigma_slope(data_slice[self.mjd_col][good], precis[good])
        result = result * 365.25 * 1e3  # Convert to mas/yr
        if (self.normalize) & (good[0].size > 0):
            new_dates = data_slice[self.mjd_col][good] * 0
            n_dates = new_dates.size
            new_dates[n_dates // 2 :] = self.baseline * 365.25
            result = (mafUtils.sigma_slope(new_dates, precis[good]) * 365.25 * 1e3) / result
        # Observations that are very close together can still fail
        if np.isnan(result):
            result = self.badval
        return result


class ParallaxCoverageMetric(BaseMetric):
    """
    Check how well the parallax factor is distributed. Subtracts the weighted mean position of the
    parallax offsets, then computes the weighted mean radius of the points.
    If points are well distributed, the mean radius will be near 1. If phase coverage is bad,
    radius will be close to zero.

    For points on the Ecliptic, uniform sampling should result in a metric value of ~0.5.
    At the poles, uniform sampling would result in a metric value of ~1.
    Conceptually, it is helpful to remember that the parallax motion of a star at the pole is
    a (nearly circular) ellipse while the motion of a star on the ecliptic is a straight line. Thus, any
    pair of observations separated by 6 months will give the full parallax range for a star on the pole
    but only observations on very specific dates will give the full range for a star on the ecliptic.

    Optionally also demand that there are observations above the snr_limit kwarg spanning theta_range radians.

    Parameters
    ----------
    m5_col: str, optional
        Column name for individual visit m5. Default fiveSigmaDepth.
    mjd_col: str, optional
        Column name for exposure time dates. Default observationStartMJD.
    filter_col: str, optional
        Column name for filter. Default filter.
    seeing_col: str, optional
        Column name for seeing (assumed FWHM). Default seeingFwhmGeom.
    rmag: float, optional
        Magnitude of fiducial star in r filter.  Other filters are scaled using sedTemplate keyword.
        Default 20.0
    sedTemplate: str, optional
        Template to use (can be 'flat' or 'O','B','A','F','G','K','M'). Default 'flat'.
    atm_err: float, optional
        Centroiding error due to atmosphere in arcsec. Default 0.01 (arcseconds).
    theta_range: float, optional
        Range of parallax offset angles to demand (in radians). Default=0 (means no range requirement).
    snr_limit: float, optional
        Only include points above the snr_limit when computing theta_range. Default 5.

    Returns
    --------
    metricValue: float
        Returns a weighted mean of the length of the parallax factor vectors.
        Values near 1 imply that the points are well distributed.
        Values near 0 imply that the parallax phase coverage is bad.
        Near the ecliptic, uniform sampling results in metric values of about 0.5.

    Notes
    -----
    Uses the ParallaxFactor stacker to calculate ra_pi_amp and dec_pi_amp.
    """

    def __init__(
        self,
        metric_name="ParallaxCoverageMetric",
        m5_col="fiveSigmaDepth",
        mjd_col="observationStartMJD",
        filter_col="filter",
        seeing_col="seeingFwhmGeom",
        rmag=20.0,
        sed_template="flat",
        atm_err=0.01,
        theta_range=0.0,
        snr_limit=5,
        **kwargs,
    ):
        cols = ["ra_pi_amp", "dec_pi_amp", m5_col, mjd_col, filter_col, seeing_col]
        units = "ratio"
        super(ParallaxCoverageMetric, self).__init__(cols, metric_name=metric_name, units=units, **kwargs)
        self.m5_col = m5_col
        self.seeing_col = seeing_col
        self.filter_col = filter_col
        self.mjd_col = mjd_col

        # Demand the range of theta values
        self.theta_range = theta_range
        self.snr_limit = snr_limit

        filters = ["u", "g", "r", "i", "z", "y"]
        self.mags = {}
        if sed_template == "flat":
            for f in filters:
                self.mags[f] = rmag
        else:
            self.mags = utils.stellarMags(sed_template, rmag=rmag)
        self.atm_err = atm_err
        caption = "Parallax factor coverage for an r=%.2f star (0 is bad, 0.5-1 is good). " % (rmag)
        caption += "One expects the parallax factor coverage to vary because stars on the ecliptic "
        caption += "can be observed when they have no parallax offset while stars at the pole are always "
        caption += "offset by the full parallax offset." ""
        self.comment = caption

    def _theta_check(self, ra_pi_amp, dec_pi_amp, snr):
        good = np.where(snr >= self.snr_limit)
        theta = np.arctan2(dec_pi_amp[good], ra_pi_amp[good])
        # Make values between 0 and 2pi
        theta = theta - np.min(theta)
        result = 0.0
        if np.max(theta) >= self.theta_range:
            # Check that things are in differnet quadrants
            theta = (theta + np.pi) % 2.0 * np.pi
            theta = theta - np.min(theta)
            if np.max(theta) >= self.theta_range:
                result = 1
        return result

    def _compute_weights(self, data_slice, snr):
        # Compute centroid uncertainty in each visit
        position_errors = mafUtils.astrom_precision(data_slice[self.seeing_col], snr, self.atm_err)
        weights = 1.0 / position_errors**2
        return weights

    def _weighted_r(self, dec_pi_amp, ra_pi_amp, weights):
        ycoord = dec_pi_amp - np.average(dec_pi_amp, weights=weights)
        xcoord = ra_pi_amp - np.average(ra_pi_amp, weights=weights)
        radius = np.sqrt(xcoord**2 + ycoord**2)
        ave_rad = np.average(radius, weights=weights)
        return ave_rad

    def run(self, data_slice, slice_point=None):
        if np.size(data_slice) < 2:
            return self.badval

        filters = np.unique(data_slice[self.filter_col])
        filters = [str(f) for f in filters]
        snr = np.zeros(len(data_slice), dtype="float")
        # compute SNR for all observations
        for filt in filters:
            in_filt = np.where(data_slice[self.filter_col] == filt)
            snr[in_filt] = mafUtils.m52snr(self.mags[str(filt)], data_slice[self.m5_col][in_filt])

        weights = self._compute_weights(data_slice, snr)
        ave_r = self._weighted_r(data_slice["ra_pi_amp"], data_slice["dec_pi_amp"], weights)
        if self.theta_range > 0:
            theta_check = self._theta_check(data_slice["ra_pi_amp"], data_slice["dec_pi_amp"], snr)
        else:
            theta_check = 1.0
        result = ave_r * theta_check
        return result


class ParallaxDcrDegenMetric(BaseMetric):
    """Use the full parallax and DCR displacement vectors to find if they are degenerate.

    Parameters
    ----------
    metricName : str, optional
        Default 'ParallaxDcrDegenMetric'.
    seeing_col : str, optional
        Default 'FWHMgeom'
    m5_col : str, optional
        Default 'fiveSigmaDepth'
    filter_col : str
        Default 'filter'
    atm_err : float
        Minimum error in photometry centroids introduced by the atmosphere (arcseconds). Default 0.01.
    rmag : float
        r-band magnitude of the fiducual star that is being used (mag).
    SedTemplate : str
        The SED template to use for fiducia star colors, passed to rubin_scheduler.utils.stellarMags.
        Default 'flat'
    tol : float
        Tolerance for how well curve_fit needs to work before believing the covariance result.
        Default 0.05.

    Returns
    -------
    metricValue : float
        Returns the correlation coefficient between the best-fit parallax amplitude and DCR amplitude.
        The RA and Dec offsets are fit simultaneously. Values close to zero are good, values close to +/- 1
        are bad. Experience with fitting Monte Carlo simulations suggests the astrometric fits start
        becoming poor around a correlation of 0.7.
    """

    def __init__(
        self,
        metric_name="ParallaxDcrDegenMetric",
        seeing_col="seeingFwhmGeom",
        m5_col="fiveSigmaDepth",
        atm_err=0.01,
        rmag=20.0,
        sed_template="flat",
        filter_col="filter",
        tol=0.05,
        **kwargs,
    ):
        self.m5_col = m5_col
        self.seeing_col = seeing_col
        self.filter_col = filter_col
        self.tol = tol
        units = "Correlation"
        # just put all the columns that all the stackers will need here?
        cols = [
            "ra_pi_amp",
            "dec_pi_amp",
            "ra_dcr_amp",
            "dec_dcr_amp",
            seeing_col,
            m5_col,
        ]
        super(ParallaxDcrDegenMetric, self).__init__(cols, metric_name=metric_name, units=units, **kwargs)
        self.mags = {}
        if sed_template == "flat":
            for f in ["u", "g", "r", "i", "z", "y"]:
                self.mags[f] = rmag
        else:
            self.mags = utils.stellarMags(sed_template, rmag=rmag)
        self.atm_err = atm_err

    def _positions(self, x, a, b):
        """
        Function to find parallax and dcr amplitudes

        x should be a vector with [[parallax_x1, parallax_x2..., parallax_y1, parallax_y2...],
        [dcr_x1, dcr_x2..., dcr_y1, dcr_y2...]]
        """
        result = a * x[0, :] + b * x[1, :]
        return result

    def run(self, data_slice, slice_point=None):
        # The idea here is that we calculate position errors (in RA and Dec) for all observations.
        # Then we generate arrays of the parallax offsets (delta RA parallax = ra_pi_amp, etc)
        #  and the DCR offsets (delta RA DCR = ra_dcr_amp, etc), and just add them together into one
        #  RA  (and Dec) offset. Then, we try to fit for how we combined these offsets, but while
        #  considering the astrometric noise. If we can figure out that we just added them together
        # (i.e. the curve_fit result is [a=1, b=1] for the function _positions above)
        # then we should be able to disentangle the parallax and DCR offsets when fitting 'for real'.
        # compute SNR for all observations
        snr = np.zeros(len(data_slice), dtype="float")
        for filt in np.unique(data_slice[self.filter_col]):
            in_filt = np.where(data_slice[self.filter_col] == filt)
            snr[in_filt] = mafUtils.m52snr(self.mags[filt], data_slice[self.m5_col][in_filt])
        # Compute the centroiding uncertainties
        # Note that these centroiding uncertainties depend on the physical size of the PSF, thus
        # we are using seeingFwhmGeom for these metrics, not seeingFwhmEff.
        position_errors = mafUtils.astrom_precision(data_slice[self.seeing_col], snr, self.atm_err)
        # Construct the vectors of RA/Dec offsets. xdata is the "input data". ydata is the "output".
        xdata = np.empty((2, data_slice.size * 2), dtype=float)
        xdata[0, :] = np.concatenate((data_slice["ra_pi_amp"], data_slice["dec_pi_amp"]))
        xdata[1, :] = np.concatenate((data_slice["ra_dcr_amp"], data_slice["dec_dcr_amp"]))
        ydata = np.sum(xdata, axis=0)
        # Use curve_fit to compute covariance between parallax and dcr amplitudes
        # Set the initial guess slightly off from the correct [1,1] to make sure it iterates.
        popt, pcov = curve_fit(
            self._positions,
            xdata,
            ydata,
            p0=[1.1, 0.9],
            sigma=np.concatenate((position_errors, position_errors)),
            absolute_sigma=True,
        )
        # Catch if the fit failed to converge on the correct solution.
        if np.max(np.abs(popt - np.array([1.0, 1.0]))) > self.tol:
            return self.badval
        # Covariance between best fit parallax amplitude and DCR amplitude.
        cov = pcov[1, 0]
        # Convert covarience between parallax and DCR amplitudes to normalized correlation
        perr = np.sqrt(np.diag(pcov))
        correlation = cov / (perr[0] * perr[1])
        result = correlation
        # This can throw infs.
        if np.isinf(result):
            result = self.badval
        return result


def calc_dist_cosines(ra1, dec1, ra2, dec2):
    # Taken from simSelfCalib.py
    """Calculates distance on a sphere using spherical law of cosines.

    Give this function RA/Dec values in radians. Returns angular distance(s), in radians.
    Note that since this is all numpy, you could input arrays of RA/Decs."""
    # This formula can have rounding errors for case where distances are small.
    # Oh, the joys of wikipedia - http://en.wikipedia.org/wiki/Great-circle_distance
    # For the purposes of these calculations, this is probably accurate enough.
    D = np.sin(dec2) * np.sin(dec1) + np.cos(dec1) * np.cos(dec2) * np.cos(ra2 - ra1)
    D = np.arccos(D)
    return D


class RadiusObsMetric(BaseMetric):
    """find the radius in the focal plane. returns things in degrees."""

    def __init__(
        self, metric_name="radiusObs", ra_col="fieldRA", dec_col="fieldDec", units="radians", **kwargs
    ):
        self.ra_col = ra_col
        self.dec_col = dec_col
        super(RadiusObsMetric, self).__init__(
            col=[self.ra_col, self.dec_col], metric_name=metric_name, units=units, **kwargs
        )

    def run(self, data_slice, slice_point):
        ra = slice_point["ra"]
        dec = slice_point["dec"]
        distances = calc_dist_cosines(
            ra,
            dec,
            np.radians(data_slice[self.ra_col]),
            np.radians(data_slice[self.dec_col]),
        )
        distances = np.degrees(distances)
        return distances

    def reduce_mean(self, distances):
        return np.mean(distances)

    def reduce_rms(self, distances):
        return np.std(distances)

    def reduce_full_range(self, distances):
        return np.max(distances) - np.min(distances)



__all__ = ("just_return", "SkyModel")

import warnings

import numpy as np
from astropy import units as u
from astropy.coordinates import AltAz, EarthLocation, SkyCoord, get_body, get_sun
from astropy.time import Time
from rubin_scheduler.utils import Site, _approx_alt_az2_ra_dec, _approx_ra_dec2_alt_az, haversine

from rubin_sim.phot_utils import Sed

from .interp_components import (
    Airglow,
    LowerAtm,
    MergedSpec,
    MoonInterp,
    ScatteredStar,
    TwilightInterp,
    UpperAtm,
    ZodiacalInterp,
)
from .utils import wrap_ra


def just_return(inval):
    """
    Really, just return the input.

    Parameters
    ----------
    input : anything

    Returns
    -------
    input : anything
        Just return whatever you sent in.
    """
    return inval


def inrange(inval, minimum=-1.0, maximum=1.0):
    """
    Make sure values are within min/max
    """
    inval = np.array(inval)
    below = np.where(inval < minimum)
    inval[below] = minimum
    above = np.where(inval > maximum)
    inval[above] = maximum
    return inval


def calc_az_rel_moon(azs, moon_az):
    az_rel_moon = wrap_ra(azs - moon_az)
    if isinstance(azs, np.ndarray):
        over = np.where(az_rel_moon > np.pi)
        az_rel_moon[over] = 2.0 * np.pi - az_rel_moon[over]
    else:
        if az_rel_moon > np.pi:
            az_rel_moon = 2.0 * np.pi - az_rel_moon
    return az_rel_moon


class SkyModel:
    def __init__(
        self,
        observatory=None,
        twilight=True,
        zodiacal=True,
        moon=True,
        airglow=True,
        lower_atm=False,
        upper_atm=False,
        scattered_star=False,
        merged_spec=True,
        mags=False,
        precise_alt_az=False,
        airmass_limit=3.0,
    ):
        """
        Instatiate the SkyModel. This loads all the required template spectra/magnitudes
        that will be used for interpolation.

        Parameters
        ----------
        Observatory : Site object
            object with attributes lat, lon, elev. But default loads LSST.

        twilight : bool (True)
            Include twilight component (True)
        zodiacal : bool (True)
            Include zodiacal light component (True)
        moon : bool (True)
            Include scattered moonlight component (True)
        airglow : bool (True)
            Include airglow component
        lower_atm : bool (False)
            Include lower atmosphere component. This component is part of `merged_spec`.
        upper_atm : bool (False)
            Include upper atmosphere component. This component is part of `merged_spec`.
        scattered_star : bool (False)
            Include scattered starlight component. This component is part of `merged_spec`.
        merged_spec : bool (True)
            Compute the lower_atm, upper_atm, and scattered_star simultaneously since they are all
            functions of only airmass.
        mags : bool (False)
            By default, the sky model computes a 17,001 element spectrum. If `mags` is True,
            the model will return the LSST ugrizy magnitudes (in that order).
        precise_alt_az : bool (False)
            If False, use the fast alt, az to ra, dec coordinate
            transformations that do not take abberation, diffraction, etc
            into account. Results in errors up to ~1.5 degrees,
            but an order of magnitude faster than coordinate transforms in sims_utils.
        airmass_limit : float (3.0)
            Most of the models are only accurate to airmass 3.0. If set higher, airmass values
            higher than 3.0 are set to 3.0.
        """

        self.moon = moon
        self.lower_atm = lower_atm
        self.twilight = twilight
        self.zodiacal = zodiacal
        self.upper_atm = upper_atm
        self.airglow = airglow
        self.scattered_star = scattered_star
        self.merged_spec = merged_spec
        self.mags = mags
        self.precise_alt_az = precise_alt_az

        # set this as a way to track if coords have been set
        self.azs = None

        # Airmass limit.
        self.airmass_limit = airmass_limit

        if self.mags:
            self.npix = 6
        else:
            self.npix = 11001

        self.components = {
            "moon": self.moon,
            "lower_atm": self.lower_atm,
            "twilight": self.twilight,
            "upper_atm": self.upper_atm,
            "airglow": self.airglow,
            "zodiacal": self.zodiacal,
            "scattered_star": self.scattered_star,
            "merged_spec": self.merged_spec,
        }

        # Check that the merged component isn't being run with other components
        merged_comps = [self.lower_atm, self.upper_atm, self.scattered_star]
        for comp in merged_comps:
            if comp & self.merged_spec:
                warnings.warn("Adding component multiple times to the final output spectra.")

        interpolators = {
            "scattered_star": ScatteredStar,
            "airglow": Airglow,
            "lower_atm": LowerAtm,
            "upper_atm": UpperAtm,
            "merged_spec": MergedSpec,
            "moon": MoonInterp,
            "zodiacal": ZodiacalInterp,
            "twilight": TwilightInterp,
        }

        # Load up the interpolation objects for each component
        self.interp_objs = {}
        for key in self.components:
            if self.components[key]:
                self.interp_objs[key] = interpolators[key](mags=self.mags)

        if observatory is None:
            self.telescope = Site("LSST")
        else:
            self.telescope = observatory
        self.location = EarthLocation(
            lat=self.telescope.latitude_rad * u.rad,
            lon=self.telescope.longitude_rad * u.rad,
            height=self.telescope.height * u.m,
        )

        # Note that observing conditions have not been set
        self.params_set = False

    def _init_points(self):
        """
        Set up an array for all the interpolation points
        """

        names = [
            "airmass",
            "nightTimes",
            "alt",
            "az",
            "azRelMoon",
            "moonSunSep",
            "moonAltitude",
            "altEclip",
            "azEclipRelSun",
            "sunAlt",
            "azRelSun",
            "solar_flux",
        ]
        types = [float] * len(names)
        self.points = np.zeros(self.npts, list(zip(names, types)))

    def set_ra_dec_mjd(
        self,
        lon,
        lat,
        mjd,
        degrees=False,
        az_alt=False,
        solar_flux=130.0,
        filter_names=["u", "g", "r", "i", "z", "y"],
    ):
        """
        Set the sky parameters by computing the sky conditions on a given MJD and sky location.



        lon: Longitude-like (RA or Azimuth). Can be single number, list, or numpy array
        lat: Latitude-like (Dec or Altitude)
        mjd: Modified Julian Date for the calculation. Must be single number.
        degrees: (False) Assumes lon and lat are radians unless degrees=True
        az_alt: (False) Assume lon, lat are RA, Dec unless az_alt=True
        solar_flux: solar flux in SFU Between 50 and 310. Default=130. 1 SFU=10^4 Jy.
        filter_names: list of fitlers to return magnitudes for (if initialized with mags=True).
        """
        self.filter_names = filter_names
        if self.mags:
            self.npix = len(self.filter_names)
        # Wrap in array just in case single points were passed
        if np.size(lon) == 1:
            lon = np.array([lon]).ravel()
            lat = np.array([lat]).ravel()
        else:
            lon = np.array(lon)
            lat = np.array(lat)
        if degrees:
            self.ra = np.radians(lon)
            self.dec = np.radians(lat)
        else:
            self.ra = lon
            self.dec = lat
        if np.size(mjd) > 1:
            raise ValueError("mjd must be single value.")
        self.mjd = mjd
        if az_alt:
            self.azs = self.ra.copy()
            self.alts = self.dec.copy()
            if self.precise_alt_az:
                raise ValueError("Need to swap in astropy")
                # self.ra, self.dec = _ra_dec_from_alt_az(
                #    self.alts,
                #    self.azs,
                #    ObservationMetaData(mjd=self.mjd, site=self.telescope),
                # )
            else:
                self.ra, self.dec = _approx_alt_az2_ra_dec(
                    self.alts,
                    self.azs,
                    self.telescope.latitude_rad,
                    self.telescope.longitude_rad,
                    mjd,
                )
        else:
            if self.precise_alt_az:
                raise ValueError("Need to swap in astropy")
                # self.alts, self.azs, pa = _alt_az_pa_from_ra_dec(
                #    self.ra,
                #    self.dec,
                #    ObservationMetaData(mjd=self.mjd, site=self.telescope),
                # )
            else:
                self.alts, self.azs = _approx_ra_dec2_alt_az(
                    self.ra,
                    self.dec,
                    self.telescope.latitude_rad,
                    self.telescope.longitude_rad,
                    mjd,
                )

        self.npts = self.ra.size
        self._init_points()

        self.solar_flux = solar_flux
        self.points["solar_flux"] = self.solar_flux

        self._setup_point_grid()

        self.params_set = True

        # Interpolate the templates to the set Parameters
        self.good_pix = np.where((self.airmass <= self.airmass_limit) & (self.airmass >= 1.0))[0]

        if self.good_pix.size <= 0:
            raise ValueError(
                "No valid points. Airmass limit=%.1f, min airmass of requested points=%.1f"
                % (self.airmass_limit, np.min(self.airmass))
            )
        else:
            self._interp_sky()

    def set_ra_dec_alt_az_mjd(
        self,
        ra,
        dec,
        alt,
        az,
        mjd,
        degrees=False,
        solar_flux=130.0,
        filter_names=["u", "g", "r", "i", "z", "y"],
    ):
        """
        Set the sky parameters by computing the sky conditions on a given MJD and sky location.

        Use if you already have alt az coordinates so you can skip the coordinate conversion.
        """
        self.filter_names = filter_names
        if self.mags:
            self.npix = len(self.filter_names)
        # Wrap in array just in case single points were passed
        if not type(ra).__module__ == np.__name__:
            if np.size(ra) == 1:
                ra = np.array([ra]).ravel()
                dec = np.array([dec]).ravel()
                alt = np.array(alt).ravel()
                az = np.array(az).ravel()
            else:
                ra = np.array(ra)
                dec = np.array(dec)
                alt = np.array(alt)
                az = np.array(az)
        if degrees:
            self.ra = np.radians(ra)
            self.dec = np.radians(dec)
            self.alts = np.radians(alt)
            self.azs = np.radians(az)
        else:
            self.ra = ra
            self.dec = dec
            self.azs = az
            self.alts = alt
        if np.size(mjd) > 1:
            raise ValueError("mjd must be single value.")
        self.mjd = mjd

        self.npts = self.ra.size
        self._init_points()

        self.solar_flux = solar_flux
        self.points["solar_flux"] = self.solar_flux

        self._setup_point_grid()

        self.params_set = True

        # Interpolate the templates to the set Parameters
        self.good_pix = np.where((self.airmass <= self.airmass_limit) & (self.airmass >= 1.0))[0]
        if self.good_pix.size <= 0:
            raise ValueError(
                "No valid points. Airmass limit=%.1f, min airmass of requested points=%.1f"
                % (self.airmass_limit, np.min(self.airmass))
            )
        else:
            self._interp_sky()

    def get_computed_vals(self):
        """
        Return the intermediate values that are caluculated by set_ra_dec_mjd and used for interpolation.
        All of these values are also accesible as class atributes, this is a convience method to grab them
        all at once and document the formats.

        Returns
        -------
        out : dict
            Dictionary of all the intermediate calculated values that may be of use outside
        (the key:values in the output dict)
        ra : numpy.array
            RA of the interpolation points (radians)
        dec : np.array
            Dec of the interpolation points (radians)
        alts : np.array
            Altitude (radians)
        azs : np.array
            Azimuth of interpolation points (radians)
        airmass : np.array
            Airmass values for each point, computed via 1./np.cos(np.pi/2.-self.alts).
        solar_flux : float
            The solar flux used (SFU).
        sunAz : float
            Azimuth of the sun (radians)
        sunAlt : float
            Altitude of the sun (radians)
        sunRA : float
            RA of the sun (radians)
        sunDec : float
            Dec of the sun (radians)
        azRelSun : np.array
            Azimuth of each point relative to the sun (0=same direction as sun) (radians)
        moonAz : float
            Azimuth of the moon (radians)
        moonAlt : float
            Altitude of the moon (radians)
        moonRA : float
            RA of the moon (radians)
        moonDec : float
            Dec of the moon (radians).  Note, if you want distances
        moon_phase : float
            Phase of the moon (0-100)
        moonSunSep : float
            Seperation of moon and sun (radians)
        azRelMoon : np.array
            Azimuth of each point relative to teh moon
        eclipLon : np.array
            Ecliptic longitude (radians) of each point
        eclipLat : np.array
            Ecliptic latitude (radians) of each point
        sunEclipLon: np.array
            Ecliptic longitude (radians) of each point with the sun at longitude zero

        Note that since the alt and az can be calculated using the fast approximation, if one wants
        to compute the distance between the the points and the sun or moon, it is probably better to
        use the ra,dec positions rather than the alt,az positions.
        """

        result = {}
        attributes = [
            "ra",
            "dec",
            "alts",
            "azs",
            "airmass",
            "solar_flux",
            "moon_phase",
            "moon_az",
            "moon_alt",
            "sun_alt",
            "sun_az",
            "az_rel_sun",
            "moon_sun_sep",
            "az_rel_moon",
            "eclip_lon",
            "eclip_lat",
            "moon_ra",
            "moon_dec",
            "sun_ra",
            "sun_dec",
            "sun_eclip_lon",
        ]

        for attribute in attributes:
            if hasattr(self, attribute):
                result[attribute] = getattr(self, attribute)
            else:
                result[attribute] = None

        return result

    def _setup_point_grid(self):
        """
        Setup the points for the interpolation functions.
        """

        time = Time(self.mjd, format="mjd")
        aa = AltAz(location=self.location, obstime=time)

        sun_coords = get_sun(time)
        self.sun_ra = sun_coords.ra.rad
        self.sun_dec = sun_coords.dec.rad

        sun_coords = sun_coords.transform_to(aa)
        self.sun_alt = sun_coords.alt.rad
        self.sun_az = sun_coords.az.rad

        # Compute airmass the same way as ESO model
        self.airmass = 1.0 / np.cos(np.pi / 2.0 - self.alts)

        self.points["airmass"] = self.airmass
        self.points["nightTimes"] = 0
        self.points["alt"] = self.alts
        self.points["az"] = self.azs

        if self.twilight:
            self.points["sunAlt"] = self.sun_alt
            self.az_rel_sun = wrap_ra(self.azs - self.sun_az)
            self.points["azRelSun"] = self.az_rel_sun

        if self.moon:
            moon_coords = get_body("moon", time)
            self.moon_ra = moon_coords.ra.rad
            self.moon_dec = moon_coords.dec.rad

            moon_coords = moon_coords.transform_to(aa)
            self.moon_alt = moon_coords.alt.rad
            self.moon_az = moon_coords.az.rad

            moon_coords = get_body("moon", time)
            sun_coords = get_sun(time)
            sep = sun_coords.separation(moon_coords)

            # looks like phase is 0-100
            self.moon_phase = sep.deg * 100 / 180.0

            # Calc azimuth relative to moon
            self.az_rel_moon = calc_az_rel_moon(self.azs, self.moon_az)
            self.moon_targ_sep = haversine(self.azs, self.alts, self.moon_az, self.moon_alt)
            # Oof, looks like some things were stored as degrees.
            self.points["moonAltitude"] += np.degrees(self.moon_alt)
            self.points["azRelMoon"] += self.az_rel_moon
            self.moon_sun_sep = sep.rad
            self.points["moonSunSep"] += np.degrees(self.moon_sun_sep)

        if self.zodiacal:
            self.eclip_lon = np.zeros(self.npts)
            self.eclip_lat = np.zeros(self.npts)

            coord = SkyCoord(ra=self.ra * u.rad, dec=self.dec * u.rad)
            coord_ecl = coord.geocentricmeanecliptic
            self.eclip_lon = coord_ecl.lon.rad
            self.eclip_lat = coord_ecl.lat.rad

            # Subtract off the sun ecliptic longitude
            sun_coords = get_sun(time)
            sun_eclip = sun_coords.geocentricmeanecliptic
            self.sun_eclip_lon = sun_eclip.lon.rad
            self.points["altEclip"] += self.eclip_lat
            self.points["azEclipRelSun"] += wrap_ra(self.eclip_lon - self.sun_eclip_lon)

        self.mask = np.where((self.airmass > self.airmass_limit) | (self.airmass < 1.0))[0]
        self.good_pix = np.where((self.airmass <= self.airmass_limit) & (self.airmass >= 1.0))[0]

    def set_params(
        self,
        airmass=1.0,
        azs=90.0,
        alts=None,
        moon_phase=31.67,
        moon_alt=45.0,
        moon_az=0.0,
        sun_alt=-12.0,
        sun_az=0.0,
        sun_eclip_lon=0.0,
        eclip_lon=135.0,
        eclip_lat=90.0,
        degrees=True,
        solar_flux=130.0,
        filter_names=["u", "g", "r", "i", "z", "y"],
    ):
        """
        Set parameters manually.
        Note, you can put in unphysical combinations of Parameters if you want to
        (e.g., put a full moon at zenith at sunset).
        if the alts kwarg is set it will override the airmass kwarg.
        MoonPhase is percent of moon illuminated (0-100)
        """

        # Convert all values to radians for internal use.
        self.filter_names = filter_names
        if self.mags:
            self.npix = len(self.filter_names)
        if degrees:
            convert_func = np.radians
        else:
            convert_func = just_return

        self.solar_flux = solar_flux
        self.sun_alt = convert_func(sun_alt)
        self.moon_phase = moon_phase
        self.moon_alt = convert_func(moon_alt)
        self.moon_az = convert_func(moon_az)
        self.eclip_lon = convert_func(eclip_lon)
        self.eclip_lat = convert_func(eclip_lat)
        self.sun_eclip_lon = convert_func(sun_eclip_lon)
        self.azs = convert_func(azs)
        if alts is not None:
            self.airmass = 1.0 / np.cos(np.pi / 2.0 - convert_func(alts))
            self.alts = convert_func(alts)
        else:
            self.airmass = airmass
            self.alts = np.pi / 2.0 - np.arccos(1.0 / airmass)
        self.moon_targ_sep = haversine(self.azs, self.alts, moon_az, self.moon_alt)
        self.npts = np.size(self.airmass)
        self._init_points()

        self.points["airmass"] = self.airmass
        self.points["nightTimes"] = 0
        self.points["alt"] = self.alts
        self.points["az"] = self.azs

        self.az_rel_moon = calc_az_rel_moon(self.azs, self.moon_az)
        self.points["moonAltitude"] += np.degrees(self.moon_alt)
        self.points["azRelMoon"] = self.az_rel_moon
        self.moon_sun_sep = self.moon_phase / 100.0 * np.pi
        self.points["moonSunSep"] += self.moon_phase / 100.0 * 180.0

        self.eclip_lon = convert_func(eclip_lon)
        self.eclip_lat = convert_func(eclip_lat)

        self.sun_eclip_lon = convert_func(sun_eclip_lon)
        self.points["altEclip"] += self.eclip_lat
        self.points["azEclipRelSun"] += wrap_ra(self.eclip_lon - self.sun_eclip_lon)

        self.sun_az = convert_func(sun_az)
        self.points["sunAlt"] = self.sun_alt
        self.points["azRelSun"] = wrap_ra(self.azs - self.sun_az)
        self.points["solar_flux"] = solar_flux

        self.params_set = True

        self.mask = np.where((self.airmass > self.airmass_limit) | (self.airmass < 1.0))[0]
        self.good_pix = np.where((self.airmass <= self.airmass_limit) & (self.airmass >= 1.0))[0]
        # Interpolate the templates to the set Parameters
        if self.good_pix.size > 0:
            self._interp_sky()
        else:
            warnings.warn("No points in interpolation range")

    def _interp_sky(self):
        """
        Interpolate the template spectra to the set RA, Dec and MJD.

        the results are stored as attributes of the class:
        .wave = the wavelength in nm
        .spec = array of spectra with units of ergs/s/cm^2/nm
        """

        if not self.params_set:
            raise ValueError(
                "No parameters have been set. Must run set_ra_dec_mjd or setParams before running interpSky."
            )

        # set up array to hold the resulting spectra for each ra, dec point.
        self.spec = np.zeros((self.npts, self.npix), dtype=float)

        # Rebuild the components dict so things can be turned on/off
        self.components = {
            "moon": self.moon,
            "lower_atm": self.lower_atm,
            "twilight": self.twilight,
            "upper_atm": self.upper_atm,
            "airglow": self.airglow,
            "zodiacal": self.zodiacal,
            "scattered_star": self.scattered_star,
            "merged_spec": self.merged_spec,
        }

        # Loop over each component and add it to the result.
        mask = np.ones(self.npts)
        for key in self.components:
            if self.components[key]:
                result = self.interp_objs[key](self.points[self.good_pix], filter_names=self.filter_names)
                # Make sure the component has something
                if np.size(result["spec"]) == 0:
                    self.spec[self.mask, :] = np.nan
                    return
                if np.max(result["spec"]) > 0:
                    mask[np.where(np.sum(result["spec"], axis=1) == 0)] = 0
                self.spec[self.good_pix] += result["spec"]
                if not hasattr(self, "wave"):
                    self.wave = result["wave"]
                else:
                    if not np.allclose(result["wave"], self.wave, rtol=1e-4, atol=1e-4):
                        warnings.warn("Wavelength arrays of components do not match.")
        if self.airmass_limit <= 2.5:
            self.spec[np.where(mask == 0), :] = 0
        self.spec[self.mask, :] = np.nan

    def return_wave_spec(self):
        """
        Return the wavelength and spectra.
        Wavelenth in nm
        spectra is flambda in ergs/cm^2/s/nm
        """
        if self.azs is None:
            raise ValueError(
                "No coordinates set. Use set_ra_dec_mjd, setRaDecAltAzMjd, or setParams methods before calling returnWaveSpec."
            )
        if self.mags:
            raise ValueError("SkyModel set to interpolate magnitudes. Initialize object with mags=False")
        # Mask out high airmass points
        # self.spec[self.mask] *= 0
        return self.wave.copy(), self.spec.copy()

    def return_mags(self, bandpasses=None):
        """
        Convert the computed spectra to a magnitude using the supplied bandpass,
        or, if self.mags=True, return the mags in the LSST filters

        Parameters
        ----------
        bandpasses : dict (None)
            Dictionary with bandpass name as keys and rubin_su=im.phot_utils.Bandpass objects as values.

        If mags=True when initialized, return mags returns an structured array with
        dtype names u,g,r,i,z,y.

        Returns
        -------
        mags : np.array
            Sky brightness in AB mags/sq arcsec
        """
        if self.azs is None:
            raise ValueError(
                "No coordinates set. Use set_ra_dec_mjd, setRaDecAltAzMjd, or setParams methods before calling return_mags."
            )

        if self.mags:
            if bandpasses:
                warnings.warn("Ignoring set bandpasses and returning LSST ugrizy.")
            mags = -2.5 * np.log10(self.spec) + np.log10(3631.0)
            # Mask out high airmass
            mags[self.mask] *= np.nan
            mags = mags.swapaxes(0, 1)
            mags_back = {}
            for i, f in enumerate(self.filter_names):
                mags_back[f] = mags[i]
        else:
            mags_back = {}
            for key in bandpasses:
                mags = np.zeros(self.npts, dtype=float) - 666
                temp_sed = Sed()
                is_through = np.where(bandpasses[key].sb > 0)
                min_wave = bandpasses[key].wavelen[is_through].min()
                max_wave = bandpasses[key].wavelen[is_through].max()
                in_band = np.where((self.wave >= min_wave) & (self.wave <= max_wave))
                for i, ra in enumerate(self.ra):
                    # Check that there is flux in the band, otherwise calc_mag fails
                    if np.max(self.spec[i, in_band]) > 0:
                        temp_sed.set_sed(self.wave, flambda=self.spec[i, :])
                        mags[i] = temp_sed.calc_mag(bandpasses[key])
                # Mask out high airmass
                mags[self.mask] *= np.nan
                mags_back[key] = mags
        return mags_back



__all__ = ("SlewContributionMetric", "AveSlewFracMetric")

import numpy as np

from .base_metric import BaseMetric

# Metrics for dealing with things from the SlewActivities table


class SlewContributionMetric(BaseMetric):
    def __init__(
        self, col="actDelay", activity=None, active_col="activity", in_crit_col="inCriticalPath", **kwargs
    ):
        """
        Return the average time, multiplied by fraction of slew --
        considering critical path activities only.
        """
        self.col = col
        self.in_crit_col = in_crit_col
        col = [col, in_crit_col]
        col.append(active_col)
        self.active_col = active_col
        self.activity = activity
        super(SlewContributionMetric, self).__init__(col=col, **kwargs)
        self.comment = "Average time for %s activity (in seconds) when in the critical path, " % (activity)
        self.comment += "multiplied by the percent of total slews in the critical path."

    def run(self, data_slice, slice_point=None):
        # Activities of this type, in critical path.
        good_in_crit = np.where(
            (data_slice[self.active_col] == self.activity) & (data_slice[self.in_crit_col] == "True")
        )[0]
        if len(good_in_crit) == 0:
            result = 0.0
        else:
            # All activities in critical path.
            in_crit = np.where((data_slice[self.in_crit_col] == "True"))[0]
            # Calculate fraction of total in-critical-path slew activities that this activity represents.
            result = np.sum(data_slice[self.col][good_in_crit]) / np.sum(data_slice[self.col][in_crit])
            #  and multiply by the mean time required by this activity.
            result *= np.mean(data_slice[self.col][good_in_crit])
        return result


class AveSlewFracMetric(BaseMetric):
    def __init__(
        self, col="actDelay", activity=None, active_col="activity", id_col="SlewHistory_slewCount", **kwargs
    ):
        """
        Return the average time multiplied by fraction of slews.
        """
        self.col = col
        self.id_col = id_col
        col = [col, id_col]
        col.append(active_col)
        self.active_col = active_col
        self.activity = activity
        super(AveSlewFracMetric, self).__init__(col=col, **kwargs)
        self.comment = "Average time for %s activity (in seconds), multiplied by percent of total slews." % (
            activity
        )

    def run(self, data_slice, slice_point=None):
        good = np.where(data_slice[self.active_col] == self.activity)[0]
        if len(good) == 0:
            result = 0.0
        else:
            result = np.mean(data_slice[self.col][good])
            nslews = np.size(np.unique(data_slice[self.id_col]))
            result = result * np.size(good) / np.float(nslews)
        return result



"""Evaluate the open shutter fraction.
"""
__all__ = ("openshutterFractions",)

import rubin_sim.maf.metric_bundles as mb
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.slicers as slicers

from .col_map_dict import col_map_dict
from .common import standard_summary


def openshutterFractions(colmap=None, runName="opsim", extraSql=None, extraInfoLabel=None):
    """Evaluate open shutter fraction over whole survey and per night.

    Parameters
    ----------
    colmap : `dict`, optional
        A dictionary with a mapping of column names.
    runName : `str`, optional
        The name of the simulated survey.
    extraSql : `str`, optional
        Additional constraint to add to any sql constraints (e.g. 'night<365')
    extraInfoLabel : `str`, optional
        Additional info_label to add before any below (i.e. "WFD").

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """
    if colmap is None:
        colmap = col_map_dict()
    bundleList = []

    group = "Open Shutter Fraction"

    subgroup = "All visits"
    if extraInfoLabel is not None:
        subgroup = extraInfoLabel + " " + subgroup.lower()
    elif extraSql is not None and extraInfoLabel is None:
        subgroup = subgroup + " " + extraSql

    # Open Shutter fraction over whole survey.
    displayDict = {"group": group, "subgroup": subgroup, "order": 0}
    displayDict["caption"] = "Total open shutter fraction over %s. " % subgroup.lower()
    displayDict["caption"] += "Does not include downtime due to weather."
    metric = metrics.OpenShutterFractionMetric(
        slew_time_col=colmap["slewtime"],
        exp_time_col=colmap["exptime"],
        visit_time_col=colmap["visittime"],
    )
    slicer = slicers.UniSlicer()
    bundle = mb.MetricBundle(metric, slicer, extraSql, info_label=subgroup, display_dict=displayDict)
    bundleList.append(bundle)
    # Count the number of nights on-sky in the survey.
    displayDict["caption"] = "Number of nights on the sky during the survey, %s." % subgroup.lower()
    metric = metrics.CountUniqueMetric(colmap["night"])
    slicer = slicers.UniSlicer()
    bundle = mb.MetricBundle(metric, slicer, extraSql, info_label=subgroup, display_dict=displayDict)
    bundleList.append(bundle)
    # Count the number of nights total in the survey
    # (start to finish of observations).
    displayDict["caption"] = "Number of nights from start to finish of survey, %s." % subgroup.lower()
    metric = metrics.FullRangeMetric(colmap["night"])
    slicer = slicers.UniSlicer()
    bundle = mb.MetricBundle(metric, slicer, extraSql, info_label=subgroup, display_dict=displayDict)
    bundleList.append(bundle)

    # Open shutter fraction per night.
    subgroup = "Per night"
    if extraInfoLabel is not None:
        subgroup = extraInfoLabel + " " + subgroup.lower()
    elif extraSql is not None and extraInfoLabel is None:
        subgroup = subgroup + " " + extraSql
    displayDict = {"group": group, "subgroup": subgroup, "order": 0}
    displayDict["caption"] = "Open shutter fraction %s." % (subgroup.lower())
    displayDict["caption"] += (
        " This compares on-sky image time against on-sky time + slews + filter "
        "changes + readout, but does not include downtime due to weather."
    )
    metric = metrics.OpenShutterFractionMetric(
        slew_time_col=colmap["slewtime"],
        exp_time_col=colmap["exptime"],
        visit_time_col=colmap["visittime"],
    )
    slicer = slicers.OneDSlicer(slice_col_name=colmap["night"], bin_size=1)
    bundle = mb.MetricBundle(
        metric,
        slicer,
        extraSql,
        info_label=subgroup,
        summary_metrics=standard_summary(),
        display_dict=displayDict,
    )
    bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(runName)
    return mb.make_bundles_dict_from_list(bundleList)



__all__ = ("DustMap3D",)

import warnings

import numpy as np

from rubin_sim.maf.maps import BaseMap
from rubin_sim.phot_utils import DustValues

from .ebv_3d_hp import ebv_3d_hp, get_x_at_nearest_y


class DustMap3D(BaseMap):
    """The DustMap3D provides a `~rubin_sim.maf.map` to hold 3d EBV data.

    Adds the following keys to the slicePoints:
    ebv3d_dists - the distances from the 3d dust map at each slice_point (in pc)
    ebv3d_ebvs - the E(B-V) values corresponding to each distance at each slice_point
    ebv3d_ebv_at_<dist_pc> - the (single) ebv value at the nearest distance to dist_pc
    ebv3d_dist_at_<d_mag> - the (single) distance value corresponding to where extinction and
    distance modulus combine to create a m-Mo value of d_mag, for the filter specified in filtername (in pc).
    Note that <dist_pc> and <d_mag> will be formatted with a single decimal place.

    The additional method 'distance_at_mag' can be called either with the distances and ebv values for the
    entire map or with the values from a single slice_point, in order to calculate the distance at which
    extinction and distance modulus combine to create a m-Mo value closest to 'dmag' in any filter.
    This is the same value as would be reported in ebv3d_dist_at_<d_mag>, but can be calculated on the fly,
    allowing variable filters and dmag values.

    Parameters
    ----------
    nside: `int`
        Healpixel resolution (2^x).
    map_file : `str`, opt
        Path to dust map file.
    interp : `bool`, opt
        Should returned values be interpolated (True) or just nearest neighbor (False).
        Default True, but is ignored if 'pixels' is provided.
    filtername : 'str', opt
        Name of the filter (to match the lsst filter names in rubin_sim.photUtils.DustValues)
        in which to calculate dust extinction magnitudes
    dist_pc : `float`, opt
        Distance at which to precalculate the nearest ebv value
    d_mag : `float`, opt
        Calculate the maximum distance which matches this 'd_mag'
        d_mag == m-mO (dust extinction + distance modulus)
    r_x : `dict` {`str`: `float`}, opt
        Per-filter dust extinction curve coefficients.
        Calculated by rubin_sim.photUtils.DustValues if "None".
    """

    def __init__(
        self,
        nside=64,
        map_file=None,
        interp=True,
        filtername="r",
        dist_pc=3000,
        d_mag=15.2,
        r_x=None,
    ):
        self.nside = nside
        self.interp = interp
        self.map_file = map_file
        self.filtername = filtername
        self.dist_pc = dist_pc
        self.d_mag = d_mag
        # r_x is the extinction coefficient (A_v = R_v * E(B-V) .. A_x = r_x * E(B-V)) per filter
        # This is equivalent to calculating A_x (using rubin_sim.photUtils.Sed.addDust) in each
        # filter and setting E(B-V) to 1 [so similar to the values calculated in DustValues ..
        # we probably should rename those (from Ax1 to r_x)
        if r_x is None:
            self.r_x = DustValues().r_x.copy()
        else:
            self.r_x = r_x
        # The values that will be added to the slicepoints
        self.keynames = [
            "ebv3d_ebvs",
            "ebv3d_dists",
            f"ebv3d_ebv_at_{self.dist_pc:.1f}",
            f"ebv3d_dist_at_{self.d_mag:.1f}",
        ]

    def run(self, slice_points):
        # If the slicer has nside, it's a healpix slicer so we can read the map directly
        if "nside" in slice_points:
            if slice_points["nside"] != self.nside:
                warnings.warn(
                    f"Slicer value of nside {slice_points['nside']} different "
                    f"from map value {self.nside}, will use correct slicer value here."
                )
            dists, ebvs = ebv_3d_hp(
                slice_points["nside"],
                pixels=slice_points["sid"],
                map_file=self.map_file,
            )
        # Not a healpix slicer, look up values based on RA,dec with possible interpolation
        else:
            dists, ebvs = ebv_3d_hp(
                self.nside,
                ra=slice_points["ra"],
                dec=slice_points["dec"],
                interp=self.interp,
                map_file=self.map_file,
            )

        # Calculate the map ebv and dist values at the initialized distance
        dist_closest, ebv_at_dist = get_x_at_nearest_y(dists, ebvs, self.dist_pc)

        # Calculate the distances at which m_minus_Mo values of 'dmag' are reached
        dist_dmag = self.distance_at_dmag(self.d_mag, dists, ebvs, self.filtername)

        slice_points["ebv3d_dists"] = dists
        slice_points["ebv3d_ebvs"] = ebvs
        slice_points[f"ebv3d_ebv_at_{self.dist_pc:.1f}"] = ebv_at_dist
        slice_points[f"ebv3d_dist_at_{self.d_mag:.1f}"] = dist_dmag

        return slice_points

    def distance_at_dmag(self, dmag, dists, ebvs, filtername=None):
        # Provide this as a method which could be used for a single slice_point as well as for whole map
        # (single slice_point means you could calculate this for any arbitrary magnitude or filter if needed)
        if filtername is None:
            filtername = self.filtername
        # calculate distance modulus for each distance
        dmods = 5.0 * np.log10(dists) - 5.0
        # calculate dust extinction at each distance, for the filtername
        a_x = self.r_x[filtername] * ebvs
        # calculate the (m-Mo) = distmod + a_x -- combination of extinction due to distance and dust
        m_minus__mo = dmods + a_x

        # Maximum distance for the given m-Mo (dmag) value
        # first do the 'within the map' closest distance
        m_minus__mo_at_mag, dist_closest = get_x_at_nearest_y(m_minus__mo, dists, dmag)
        # calculate distance modulus for an object with the maximum dust extinction (and then the distance)
        if a_x.ndim == 2:
            dist_mods_far = dmag - a_x[:, -1]
        else:
            dist_mods_far = dmag - a_x.max()
        dists_far = 10.0 ** (0.2 * dist_mods_far + 1.0)
        # Use furthest of the two
        dist_dmag = np.where(dists_far > dist_closest, dists_far, dist_closest)
        return dist_dmag



__all__ = ("generate_catalog",)

import sys
import warnings

import numpy as np
import numpy.lib.recfunctions as rfn
from scipy.spatial import cKDTree as kdtree

from .offsets import OffsetSNR
from .star_tools import assign_patches, stars_project


def wrap_ra(ra):
    """Wraps RA into 0-360 degrees."""
    ra = ra % 360.0
    return ra


def cap_dec(dec):
    """Terminates declination at +/- 90 degrees."""
    dec = np.where(dec > 90, 90, dec)
    dec = np.where(dec < -90, -90, dec)
    return dec


def treexyz(ra, dec):
    """Calculate x/y/z values for ra/dec points, ra/dec in radians."""
    # Note ra/dec can be arrays.
    x = np.cos(dec) * np.cos(ra)
    y = np.cos(dec) * np.sin(ra)
    z = np.sin(dec)
    return x, y, z


def build_tree(ra, dec, leafsize=100):
    """Build KD tree on RA/dec and set radius (via setRad) for matching.

    ra, dec = RA and Dec values (in radians).
    leafsize = the number of Ra/Dec pointings in each leaf node."""
    if np.any(np.abs(ra) > np.pi * 2.0) or np.any(np.abs(dec) > np.pi * 2.0):
        raise ValueError("Expecting RA and Dec values to be in radians.")
    x, y, z = treexyz(ra, dec)
    data = list(zip(x, y, z))
    if np.size(data) > 0:
        star_tree = kdtree(data, leafsize=leafsize)
    else:
        raise ValueError("ra and dec should have length greater than 0.")
    return star_tree


def generate_catalog(
    visits,
    stars_array,
    offsets=None,
    lsst_filter="r",
    n_patches=16,
    radius_fov=1.8,
    seed=42,
    uncert_floor=0.005,
    verbose=True,
):
    """
    Generate a catalog of observed stellar magnitudes.

    Parameters
    ----------
    visits : np.array
        A numpy array with the properties of the visits.
        Expected columns of fiveSigmaDepth, ra, dec, rotSkyPos (all degrees)
    offsets : list of rubin_sim.selfcal.Offset classes
        A list of instatiated classes that will apply offsets to the stars
    lsst_filter :  str ("r")
        Which filter to use for the observed stars.
    n_patches : int (16)
        Number of patches to divide the FoV into.  Must be an integer squared
    radius_fov : float (1.8)
        Radius of the telescope field of view in degrees
    seed : float (42)
        Random number seed
    uncert_floor : float (0.005)
        Value to add in quadrature to magnitude uncertainties (mags)
    verbose : bool (True)
        Should we be verbose
    """

    if offsets is None:
        # Maybe change this to just run with a default SNR offset
        warnings.warn("Warning, no offsets set, returning without running")
        return

    # For computing what the 'expected' uncertainty on the observation will be
    mag_uncert = OffsetSNR(lsst_filter=lsst_filter)

    # set the radius for the kdtree
    x0, y0, z0 = (1, 0, 0)
    x1, y1, z1 = treexyz(np.radians(radius_fov), 0)
    tree_radius = np.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2 + (z1 - z0) ** 2)

    newcols = [
        "x",
        "y",
        "radius",
        "patch_id",
        "sub_patch",
        "observed_mag",
        "mag_uncert",
    ]
    newtypes = [float, float, float, int, int, float, float]
    stars_new = np.zeros(stars_array.size, dtype=list(zip(newcols, newtypes)))

    # only need to output these columns (saving rmag+ for now for convienence)
    output_cols = [
        "id",
        "patch_id",
        "observed_mag",
        "mag_uncert",
        "%smag" % lsst_filter,
        "ra",
        "decl",
    ]
    output_dtypes = [int, int, float, float, float, float, float]

    stars = rfn.merge_arrays([stars_array, stars_new], flatten=True)

    # Build a KDTree for the stars
    star_tree = build_tree(np.radians(stars["ra"]), np.radians(stars["decl"]))

    # XXX--maybe update the way seeding is going on
    np.random.seed(seed)

    list_of_observed_arrays = []

    n_visits = np.size(visits)

    for i, visit in enumerate(visits):
        dmags = {}
        # Calc x,y, radius for each star, crop off stars outside the FoV
        # XXX - plan to replace with code to see where each star falls and get chipID.
        vx, vy, vz = treexyz(np.radians(visit["ra"]), np.radians(visit["dec"]))
        indices = star_tree.query_ball_point((vx, vy, vz), tree_radius)
        stars_in = stars[indices]
        stars_in = stars_project(stars_in, visit)

        # Assign patchIDs
        stars_in = assign_patches(stars_in, visit, n_patches=n_patches)

        # Apply the offsets that have been configured
        for offset in offsets:
            dmags[offset.newkey] = offset(stars_in, visit, dmags=dmags)

        # Total up all the dmag's to make the observed magnitude
        keys = list(dmags.keys())
        obs_mag = stars_in["%smag" % lsst_filter].copy()
        for key in keys:
            obs_mag += dmags[key]

        # Calculate the uncertainty in the observed mag:
        mag_err = (
            mag_uncert.calc_mag_errors(obs_mag, err_only=True, m5=visit["fiveSigmaDepth"]) ** 2
            + uncert_floor**2
        ) ** 0.5

        # put values into the right columns
        stars_in["observed_mag"] = obs_mag
        stars_in["mag_uncert"] = mag_err
        # Should shrink this down so we only return the needed columns
        # observed_mag, mag_uncert, patchid, star_id
        sub_cols = np.empty(stars_in.size, dtype=list(zip(output_cols, output_dtypes)))
        for key in output_cols:
            sub_cols[key] = stars_in[key]
        list_of_observed_arrays.append(sub_cols)
        if verbose:
            progress = i / n_visits * 100
            text = "\rprogress = %.2f%%" % progress
            sys.stdout.write(text)
            sys.stdout.flush()

    result = np.concatenate(list_of_observed_arrays)
    return result
